import json

from tests import BaseTestCase


class TestHttpMocked(BaseTestCase):

    def setUp(self) -> None:
        super(TestHttpMocked, self).setUp()

    def test_mocked_me(self):
        self.mock_request.return_value = {'data': {'me': {'foo': 'barbar'}}}
        self.assertEqual(dict(foo='barbar'), json.loads(self.cli_stdout(['app.py', 'me'])))


class TestStdoutStderrSideEffect(BaseTestCase):

    def setUp(self) -> None:
        super(TestStdoutStderrSideEffect, self).setUp()

    def test_stdout_stderr_side_effect_removed(self):
        self.mock_request.return_value = {'data': {'me': {'message': 'side-effect'}}}

        output = self.cli_stdout(['app.py', 'me'])
        self.assertEqual({'message': 'side-effect'}, json.loads(output))

        output = self.cli_stdout(['app.py', 'me'])
        self.assertEqual({'message': 'side-effect'}, json.loads(output))

        help1 = self.cli_stderr(['app.py'])
        help2 = self.cli_stderr(['app.py'])
        self.assertEqual(help1, help2)



from primehub import Helpful, cmd, Module
from primehub.utils import create_logger, group_not_found

logger = create_logger('cmd-config')


class Config(Helpful, Module):
    """
    The config module provides functions to update PrimeHubConfig.

    Config class only changes data in the memory,
    otherwise CliConfig will change data and update it to the real configuration file.

    In SDK mode, config will use Config.
    In Command mode, config will use CliConfig.
    CliConfig will update the file config after a `set_*` method call.
    """

    def set_endpoint(self, endpoint: str):
        """
        Set endpoint to the GraphQL API.

        It usually is in this pattern https://<primehub-domain>/api/graphql

        :type endpoint: str
        :param endpoint: an URL to GraphQL API endpoint
        """
        self.primehub.primehub_config.endpoint = endpoint

    def set_token(self, token: str):
        """
        Set api-token to the PrimeHubConfig

        :type token: str
        :param token: a token used by GraphQL request
        """
        self.primehub.primehub_config.api_token = token

    def set_group(self, group: str):
        """
        Set current group to work with group aware actions.

        When setting a group, it will validate the group. It the group is invalid, PrimeHub.current_group will be None.

        :type group: str
        :param group: group name
        """
        self.reconfigure_group(group, raise_error=True)

    def reconfigure_group(self, group, raise_error=False):
        if group is None:
            return
        selected_group = self._fetch_group_info(group)
        if selected_group is None:
            if raise_error:
                logger.warning(
                    '[reconfigure_group] Cannot find the group [%s] in the effective groups, '
                    'keep the original group [%s]',
                    group, self.primehub.primehub_config.current_group.get('name'))
                group_not_found(group)
        else:
            self.primehub.primehub_config.current_group = selected_group

    def _fetch_group_info(self, group):
        query = """
        {
          me {
            effectiveGroups {
              id
              name
              displayName
            }
          }
        }
        """
        results = self.request({}, query)
        for g in results['data']['me']['effectiveGroups']:
            if g['name'] == group:
                return g

    def help_description(self):
        return "Update the settings of PrimeHub SDK"


class CliConfig(Config):

    @cmd(name='set-endpoint', description='set endpoint and save to the config file')
    def set_endpoint(self, endpoint: str):
        """
        Set endpoint to the GraphQL API.

        It usually is in this pattern https://<primehub-domain>/api/graphql

        :type endpoint: str
        :param endpoint: an URL to GraphQL API endpoint
        """
        super(CliConfig, self).set_endpoint(endpoint)
        self.update()

    @cmd(name='set-token', description='set token and save to the config file')
    def set_token(self, token: str):
        """
        Set api-token to the PrimeHubConfig

        :type token: str
        :param token: a token used by GraphQL request
        """
        super(CliConfig, self).set_token(token)
        self.update()

    @cmd(name='set-group', description='set group and save to the config file')
    def set_group(self, group: str):
        """
        Set current group to work with group aware actions.

        When setting a group, it will validate the group. It the group is invalid, PrimeHub.current_group will be None.

        :type group: str
        :param group: group name
        """
        super(CliConfig, self).set_group(group)
        if not self.primehub.primehub_config.group_info:
            group_not_found(group)
        if not self.primehub.primehub_config.group_info.get('id', None):
            group_not_found(group)
        self.primehub.primehub_config.group = group
        self.update()

    def update(self):
        self.primehub.primehub_config.save()



import functools
import io
import json
import os
from unittest import TestCase, mock

from primehub import PrimeHub, PrimeHubConfig, cli
from primehub.utils import create_logger

logger = create_logger('primehub-test')


def reset_stderr_stdout(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        finally:
            args[0].reset_cli_output()

    return wrapper


class BaseTestCase(TestCase):
    """
    A base test case which mocks out the low-level HTTP Client to prevent
    any actual calls to the API.
    """

    def setUp(self) -> None:

        import primehub.utils.http_client

        primehub.utils.http_client.Client.request = mock.MagicMock()
        self.mock_request = primehub.utils.http_client.Client.request

        import primehub
        self.test_default_config_path = self.tempfile()
        primehub.PrimeHubConfig.get_default_path = mock.MagicMock()
        primehub.PrimeHubConfig.get_default_path.return_value = self.test_default_config_path
        print("the default path is ", self.test_default_config_path)

        self.sdk = PrimeHub(PrimeHubConfig())
        self.sdk.stderr = io.StringIO()
        self.sdk.stdout = io.StringIO()

    def tearDown(self) -> None:
        os.environ['PRIMEHUB_GROUP'] = ""
        os.environ['PRIMEHUB_API_TOKEN'] = ""
        os.environ['PRIMEHUB_API_ENDPOINT'] = ""

    @reset_stderr_stdout
    def cli_stderr(self, argv: list):
        print("cli_stderr", argv)
        self.invoke_cli(argv)
        return self.sdk.stderr.getvalue()

    @reset_stderr_stdout
    def cli_stdout(self, argv: list):
        print("cli_stdout", argv)
        self.invoke_cli(argv)
        return self.sdk.stdout.getvalue()

    def invoke_cli(self, argv: list):
        try:
            import sys
            sys.argv = argv
            if [x for x in argv if not isinstance(x, str)]:
                raise ValueError('all arguments must be a str type')
            if '--json' not in sys.argv:
                sys.argv.append('--json')
            cli.main(sdk=self.sdk)
        except SystemExit:
            pass

    def reset_cli_output(self):
        logger.info("\n{} reset_cli_output {}".format("=" * 40, "=" * 40))
        logger.info("\nSTDOUT:\n\n{}\n\nSTDERR:\n\n{}\n\n".format(self.sdk.stdout.getvalue(),
                                                                  self.sdk.stderr.getvalue()))
        self.sdk.stderr = io.StringIO()
        self.sdk.stdout = io.StringIO()

    def create_fake_config(self, content: dict) -> str:
        from tempfile import mkstemp
        fd, p = mkstemp('.json', text=True)
        with open(p, "w") as fh:
            fh.write(json.dumps(content))
        return p

    def create_config_dict(self, endpoint, token, group):
        return {'endpoint': endpoint, 'api-token': token, 'group': dict(name=group)}

    def cfg_dict_with_prefix(self, prefix: str):
        return self.create_config_dict(prefix + ":endpoint", prefix + ":api-token", prefix + ":group")

    def assert_config_with_prefix(self, prefix: str, cfg: PrimeHubConfig):
        self.assertEqual(prefix + ':endpoint', cfg.endpoint)
        self.assertEqual(prefix + ':api-token', cfg.api_token)
        self.assertEqual(cfg.current_group.get('name'), cfg.group)
        self.assertEqual(prefix + ':group', cfg.current_group.get('name'))

    def tempfile(self):
        from tempfile import mkstemp
        fd, p = mkstemp(".data", text=True)
        return p

    def load_json_file(self, path):
        with open(path, "r") as fh:
            return json.load(fh)



from typing import Optional

from primehub import Helpful, cmd, Module
from primehub.utils import resource_not_found


class Groups(Helpful, Module):
    """
    List effective groups or get a group entry from the list
    """

    @cmd(name='list', description='List groups')
    def list(self) -> list:
        """
        List available groups

        :rtype: list
        :returns: all effective groups for your account
        """
        query = """
        {
          me {
            effectiveGroups {
              id
              name
              displayName
              # user quota
              quotaCpu
              quotaGpu
              quotaMemory
              # group quota
              projectQuotaCpu
              projectQuotaGpu
              projectQuotaMemory
            }
          }
        }
        """
        results = self.request({}, query)
        return results['data']['me']['effectiveGroups']

    @cmd(name='get', description='Get group by name', return_required=True)
    def get(self, group_name: str) -> Optional[dict]:
        """
        Get a group from available groups

        :type group_name: str
        :param group_name: the name of a group

        :rtype: Optional[dict]
        :returns: a group entry from available groups
        """
        groups = self.list()
        group = [x for x in groups if x['name'] == group_name]
        if group:
            return group[0]

        resource_not_found('group', group_name, 'name')
        return None

    def help_description(self):
        return "Get a group or list groups"



from typing import Optional

from primehub import Helpful, cmd, Module
from primehub.resource_operations import GroupResourceOperation


class Datasets(Helpful, Module, GroupResourceOperation):
    """
    List datasets or get a dataset entry from the list
    """
    resource_name = 'datasets'
    query = """
    query {
      me {
        effectiveGroups {
          name
          datasets {
            id
            name
            displayName
            description
            type
          }
        }
      }
    }
    """

    @cmd(name='list', description='List datasets')
    def list(self) -> list:
        """
        List datasets

        :rtype: list
        :returns: all datasets in the current group
        """
        return self.do_list(Datasets.query, Datasets.resource_name)

    @cmd(name='get', description='Get a dataset by name', return_required=True)
    def get(self, name) -> Optional[dict]:
        """
        Get a dataset from the current group

        :type name: str
        :param name: the name of a dataset

        :rtype: Optional[dict]
        :returns: a dataset
        """
        return self.do_get(Datasets.query, Datasets.resource_name, name)

    def help_description(self):
        return "Get a dataset or list datasets"



import json

from types import GeneratorType
from typing import Any, TextIO, Union, Dict, List

from tabulate import tabulate

from primehub import create_logger
from abc import ABCMeta, abstractmethod

logger = create_logger('display')

CUSTOMIZED_COLUMNS: Dict[str, List[tuple]] = {
    'primehub.jobs.list': [
        ('id',),
        ('displayName',),
        ('schedule',),
        ('userName', 'user',),
        ('startTime',),
        ('finishTime',),
        ('phase',),
    ],
    'primehub.schedules.list': [
        ('id',),
        ('displayName',),
        ('recurrence',),
        ('userName', 'createdBy',),
        ('nextRunTime',),
    ],
    'primehub.images.list': [
        ('name',),
        ('displayName',),
        ('description',),
    ],
    'primehub.deployments.list': [
        ('id',),
        ('name',),
        ('modelImage',),
        ('description',),
        ('stop',),
        ('status',),
        ('message',),
    ]
}


def display_tree_like_format(data, file: TextIO, width=0, indent=0):
    output = data.items()
    if width == 0:
        width = 2 + max([len(x[0]) for x in output])
        if width < 15:
            width = 15

    for k, v in output:
        remaining = width - len(k)
        if isinstance(v, dict):
            print("{}{}:".format(" " * indent, k), file=file)
            display_tree_like_format(v, file, width - 2, indent + 2)
        else:
            print("{}{}:{}{}".format(" " * indent, k, " " * remaining, v), file=file)


class Displayable(metaclass=ABCMeta):

    @property
    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def display(self, action: dict, value: Any, file: TextIO):
        pass

    @abstractmethod
    def display_many(self, action: dict, value: Union[dict, list], file: TextIO):
        pass

    @abstractmethod
    def display_single(self, action: dict, value: Any, file: TextIO):
        pass


def wrapper_generator(gen):
    is_type_sent = False
    for x in gen:
        if not is_type_sent:
            yield isinstance(x, (str, bytes))
            is_type_sent = True
        yield x


class Display(Displayable):

    @property
    def name(self):
        return 'json'

    def display(self, action: dict, value: Any, file: TextIO):
        if not value:
            return
        logger.debug('use display %s', self.name)

        if isinstance(value, GeneratorType):
            logger.debug('display generator-type')
            wrapped_generator = wrapper_generator(value)
            is_str = next(wrapped_generator, None)
            if is_str is None:
                return
            if is_str:
                for x in wrapped_generator:
                    self.display_single(action, x, file)
            else:
                self.display(action, list(wrapped_generator), file)
        else:
            if isinstance(value, dict) or isinstance(value, list):
                self.display_many(action, value, file)
            else:
                self.display_single(action, value, file)

    def display_many(self, action: dict, value: Union[dict, list], file: TextIO):
        logger.debug('display-many')
        json.dump(value, file)

    def display_single(self, action: dict, value: Any, file: TextIO):
        logger.debug('display-single')
        if isinstance(value, str):
            print(value, file=file)
        elif isinstance(value, bytes):
            file.write(value.decode())
        else:
            json.dump(value, file)


class HumanFriendlyDisplay(Displayable):

    @property
    def name(self):
        return 'human-friendly'

    def display(self, action: dict, value: Any, file: TextIO):
        if not value:
            return
        logger.debug('use display %s', self.name)

        if isinstance(value, GeneratorType):
            logger.debug('display generator-type')
            wrapped_generator = wrapper_generator(value)
            is_str = next(wrapped_generator, None)
            if is_str is None:
                return
            if is_str:
                for x in wrapped_generator:
                    self.display_single(action, x, file)
            else:
                self.display(action, list(wrapped_generator), file)
        else:
            # for human-friendly dict type will use display_single
            if isinstance(value, list):
                self.display_many(action, value, file)
            else:
                self.display_single(action, value, file)

    def display_many(self, action: dict, value: Union[dict, list], file: TextIO):
        logger.debug('display-many')

        # 'module': 'primehub.jobs', 'func': 'list'
        customized_key = "{}.{}".format(action['module'], action['func'])
        if customized_key not in CUSTOMIZED_COLUMNS:
            print(tabulate(value, headers="keys"), file=file)
        else:
            def transform(columns: list, v: Union[dict, list]):
                for x in v:
                    result = {}
                    for c in columns:
                        column_name = column_alias = c[0]
                        if len(c) == 2:
                            column_alias = c[1]
                        result[column_alias] = x[column_name]
                    yield result

            print(tabulate(transform(CUSTOMIZED_COLUMNS[customized_key], value), headers="keys"), file=file)

    def display_single(self, action: dict, value: Any, file: TextIO):
        logger.debug('display-single')
        if isinstance(value, str):
            print(value, file=file)
        elif isinstance(value, bytes):
            file.write(value.decode())
        else:
            display_tree_like_format(value, file)



from typing import Iterator

from primehub import Helpful, cmd, Module
import urllib.parse

from primehub.utils.optionals import toggle_flag


class Notebooks(Helpful, Module):
    """
    The notebooks module provides functions to manage Primehub Notebooks
    """

    @cmd(name='logs', description='Get notebooks logs', optionals=[('follow', toggle_flag), ('tail', int)])
    def logs(self, **kwargs) -> Iterator[str]:
        """
        Get notebooks logs

        :type follow: bool
        :param follow: Wait for additional logs to be appended

        :type tail: int
        :param tail: Show last n lines

        :rtype str
        :return The notebooks logs
        """
        follow = kwargs.get('follow', False)
        tail = kwargs.get('tail', 10)

        endpoint = urllib.parse.urljoin(self.endpoint, '/api/logs/jupyterhub')
        return self.primehub.request_logs(endpoint, follow, tail)

    def help_description(self):
        return "Get notebooks logs"



import os
from tempfile import mkstemp

from primehub import PrimeHubConfig
from tests import BaseTestCase


class TestPrimeHubConfig(BaseTestCase):

    def setUp(self) -> None:
        super(TestPrimeHubConfig, self).setUp()

    def test_primehub_config_evaluation(self):
        endpoint = 'https://example.primehub.io/api/graphql'
        token = 'my-token'
        group = 'my-phusers'

        raw_config = self.create_config_dict(endpoint, token, group)
        default_cfg_path = self.create_fake_config(raw_config)

        class MyPrimeHubConfig(PrimeHubConfig):
            def get_default_path(self):
                return default_cfg_path

        # Verify config from the default path
        cfg = MyPrimeHubConfig()
        self.assertEqual(endpoint, cfg.endpoint)
        self.assertEqual(token, cfg.api_token)
        self.assertEqual(group, cfg.group)

        # Verify config from alternative path
        raw_config['endpoint'] = 'https://different-path.primehub.io'
        alternative_config_file = self.create_fake_config(raw_config)
        cfg = MyPrimeHubConfig(config=alternative_config_file)
        self.assertEqual('https://different-path.primehub.io', cfg.endpoint)
        self.assertEqual(cfg.get_default_path(), default_cfg_path)
        self.assertEqual(cfg.config_file, alternative_config_file)

        # Verify config override by ENV
        os.environ['PRIMEHUB_GROUP'] = 'env:group'
        os.environ['PRIMEHUB_API_TOKEN'] = 'env:api-token'
        os.environ['PRIMEHUB_API_ENDPOINT'] = 'env:endpoint'

        cfg.load_config_from_env()  # force reload env-config
        self.assert_config_with_prefix('env', cfg)

        # Verify override from user
        cfg.endpoint = 'user:endpoint'
        cfg.api_token = 'user:api-token'
        cfg.group = 'user:group'
        self.assert_config_with_prefix('user', cfg)

    def test_primehub_config_save(self):
        endpoint = 'file:endpoint'
        token = 'file:api-token'
        group = 'file:group'

        # Arrange the original configuration file
        raw_config = self.create_config_dict(endpoint, token, group)
        config_file = self.create_fake_config(raw_config)
        cfg = PrimeHubConfig(config=config_file)
        self.assert_config_with_prefix('file', cfg)

        # Assume the user updating the configuration and persist it to the file
        cfg.endpoint = 'save:endpoint'
        cfg.api_token = 'save:api-token'
        cfg.group = 'save:group'

        fd, new_config_path = mkstemp('.json', text=True)
        cfg.save(new_config_path)

        # Verify the new configuration becoming with `save:*` values
        self.assert_config_with_prefix('save', cfg)

    def test_set_from_constructor(self):
        endpoint = 'file:endpoint'
        token = 'file:api-token'
        group = 'file:group'

        # Arrange the original configuration file
        raw_config = self.create_config_dict(endpoint, token, group)
        config_file = self.create_fake_config(raw_config)
        cfg = PrimeHubConfig(config=config_file, endpoint='prop:endpoint')
        self.assertEqual('prop:endpoint', cfg.endpoint)
        self.assertEqual('file:api-token', cfg.api_token)
        self.assertEqual('file:group', cfg.group)



import os
from unittest import mock

from primehub.files import _normalize_user_input_path, _normalize_dest_path
from tests import BaseTestCase
from primehub.utils import create_logger, SharedFileException

logger = create_logger('primehub-test')

"""
Consider folloing files
PHFS:
/
├── l0.csv
└── deep/
    ├── l1.csv
    └── sub/
        ├── l2.csv
        └── path/
            └── l3.csv

Current local directroy:
./
├── sub (file)
└── l2.csv/
"""


def mock_request_side_effect(*args):
    phfs_files = ['/l0.csv', '/deep/l1.csv', '/deep/sub/l2.csv', '/deep/sub/path/l3.csv']

    prefix = args[0]['where']['phfsPrefix']
    recursive = args[0]['options']['recursive']
    limit = args[0]['options']['limit']

    filter_files = [f[len(prefix):] for f in phfs_files if f.startswith(prefix)]
    if not recursive:
        if prefix == '/':
            filter_files = [f for f in filter_files if f]
        else:
            filter_files = [f[1:] for f in filter_files if f and f[0] == '/']
        filter_files = [f if f.find('/') == -1 else f[:f.find('/') + 1] for f in filter_files]

    filter_files = list(set(filter_files))
    filter_files = filter_files[: limit]
    items = [{'name': f} for f in filter_files]
    return {'data': {'files': {'items': items}}}


def mock_walk_side_effect(name):
    if name not in ['.', './', '']:
        return []
    else:
        return [('./', ('l2.csv',), ('sub',)), ('./l2.csv', (), ())]


def mock_isfile_side_effect(name):
    return name in ['./sub', 'sub']


def mock_isdir_side_effect(name):
    return name in ['/', '.', './', './l2.csv', './l2.csv/', 'l2.csv', 'l2.csv/']


class TestCmdFiles(BaseTestCase):
    def setUp(self) -> None:
        super(TestCmdFiles, self).setUp()
        self.maxDiff = None

        self.sdk.primehub_config.endpoint = 'https://example.primehub.io/api/graphql'
        self.sdk.primehub_config.group_info = {'name': 'phusers', 'id': 'any-id'}
        self.mock_request.return_value = {
            'data': {
                'me': {
                    'effectiveGroups': [
                        {'name': 'phusers', 'foobar': [dict(name="3", value="3")]},
                    ]
                }
            }
        }

        self.mock_request.side_effect = mock_request_side_effect

    def assertFileListEqual(self, a: list, b: list):
        self.assertEqual(sorted(a), sorted(b))

    def assertDictListEqual(self, a: list, b: list):
        from operator import itemgetter
        a = sorted(a, key=itemgetter('name'))
        b = sorted(b, key=itemgetter('name'))
        self.assertEqual(a, b)

    def test_files_list(self):
        list_cmd = self.sdk.files.list

        # single file
        self.assertDictListEqual(list_cmd('/l0.csv'), [{'name': 'l0.csv'}])
        self.assertDictListEqual(list_cmd('/deep/sub/path/l3.csv'), [{'name': 'l3.csv'}])

        # directory
        self.assertDictListEqual(list_cmd('/'), [{'name': 'deep/'}, {'name': 'l0.csv'}])
        self.assertDictListEqual(list_cmd('/deep'), [{'name': 'sub/'}, {'name': 'l1.csv'}])
        self.assertDictListEqual(list_cmd('/deep/'), [{'name': 'sub/'}, {'name': 'l1.csv'}])
        self.assertDictListEqual(list_cmd('deep'), [{'name': 'sub/'}, {'name': 'l1.csv'}])

        # incomplete or non-exist path
        with self.assertRaises(SharedFileException) as e:
            self.assertDictListEqual(list_cmd('/l0'), [])
        self.assertEqual(str(e.exception), 'No such file or directory: /l0')

        with self.assertRaises(SharedFileException) as e:
            self.assertDictListEqual(list_cmd('/l0.csv/'), [])
        self.assertEqual(str(e.exception), 'Not a directory: /l0.csv/')

        with self.assertRaises(SharedFileException) as e:
            self.assertDictListEqual(list_cmd('/de'), [])
        self.assertEqual(str(e.exception), 'No such file or directory: /de')

        with self.assertRaises(SharedFileException) as e:
            self.assertDictListEqual(list_cmd('/none'), [])
        self.assertEqual(str(e.exception), 'No such file or directory: /none')

    @mock.patch('os.walk', mock.MagicMock(side_effect=mock_walk_side_effect))
    @mock.patch('os.path.isdir', mock.MagicMock(side_effect=mock_isdir_side_effect))
    @mock.patch('os.path.isfile', mock.MagicMock(side_effect=mock_isfile_side_effect))
    def test_files_get_download_src_dst_list_directory_1(self):
        get_download_src_dst_list = self.sdk.files._generate_download_list
        expected = [('/l0.csv', 'l0.csv'),
                    ('/deep/l1.csv', 'deep/l1.csv'),
                    ('/deep/sub/l2.csv', 'deep/sub/l2.csv'),
                    ('/deep/sub/path/l3.csv', 'deep/sub/path/l3.csv')]

        actual = get_download_src_dst_list('/', '.', recursive=True)
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, expected)

        # directory has to downlowd recursively
        with self.assertRaises(SharedFileException):
            actual = get_download_src_dst_list('.', '.')

    @mock.patch('os.walk', mock.MagicMock(side_effect=mock_walk_side_effect))
    @mock.patch('os.path.isdir', mock.MagicMock(side_effect=mock_isdir_side_effect))
    @mock.patch('os.path.isfile', mock.MagicMock(side_effect=mock_isfile_side_effect))
    def test_files_get_download_src_dst_list_directory_2(self):
        get_download_src_dst_list = self.sdk.files._generate_download_list

        actual = get_download_src_dst_list('/deep', '.', recursive=True)
        actual = self.norm_src_dst_list(actual)
        expected = [('/deep/l1.csv', 'deep/l1.csv'),
                    ('/deep/sub/l2.csv', 'deep/sub/l2.csv'),
                    ('/deep/sub/path/l3.csv', 'deep/sub/path/l3.csv')]
        self.assertFileListEqual(actual, expected)

        actual = get_download_src_dst_list('/deep/sub', '', recursive=True)
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [])

        with self.assertRaises(SharedFileException) as e:
            actual = get_download_src_dst_list('/deep', './sub', recursive=True)
        self.assertEqual(str(e.exception), 'Not a directory: ./sub')

        with self.assertRaises(SharedFileException) as e:
            actual = get_download_src_dst_list('/deep', 'not/exist/dir', recursive=True)
        self.assertEqual(str(e.exception), 'No such file or directory: not/exist')

        actual = get_download_src_dst_list('/deep/', '.', recursive=True)
        actual = self.norm_src_dst_list(actual)
        expected = [('/deep/l1.csv', 'l1.csv')]
        self.assertFileListEqual(actual, expected)

        actual = get_download_src_dst_list('/deep/sub/', '.', recursive=True)
        actual = self.norm_src_dst_list(actual)
        expected = [('/deep/sub/path/l3.csv', 'path/l3.csv')]
        self.assertFileListEqual(actual, expected)

        actual = get_download_src_dst_list('/deep/sub/', 'new', recursive=True)
        actual = self.norm_src_dst_list(actual)
        expected = [('/deep/sub/l2.csv', 'new/l2.csv'),
                    ('/deep/sub/path/l3.csv', 'new/path/l3.csv')]

        self.assertFileListEqual(actual, expected)
        actual = get_download_src_dst_list('/deep/sub', 'new', recursive=True)
        actual = self.norm_src_dst_list(actual)
        expected = [('/deep/sub/l2.csv', 'new/sub/l2.csv'),
                    ('/deep/sub/path/l3.csv', 'new/sub/path/l3.csv')]
        self.assertFileListEqual(actual, expected)

    @mock.patch('os.walk', mock.MagicMock(side_effect=mock_walk_side_effect))
    @mock.patch('os.path.isdir', mock.MagicMock(side_effect=mock_isdir_side_effect))
    @mock.patch('os.path.isfile', mock.MagicMock(side_effect=mock_isfile_side_effect))
    def test_files_get_download_src_dst_list_file(self):
        get_download_src_dst_list = self.sdk.files._generate_download_list
        # recursive does not effect donwload a file
        actual = get_download_src_dst_list('l0.csv', '.', recursive=True)
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [('/l0.csv', 'l0.csv')])

        actual = get_download_src_dst_list('l0.csv', '.')
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [('/l0.csv', 'l0.csv')])

        actual = get_download_src_dst_list('l0.csv', '/')
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [('/l0.csv', '/l0.csv')])

        actual = get_download_src_dst_list('l0.csv', 'sub')
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [('/l0.csv', 'sub')])

        actual = get_download_src_dst_list('l0.csv', 'l2.csv')
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [('/l0.csv', 'l2.csv/l0.csv')])

        actual = get_download_src_dst_list('l0.csv', 'l2.csv/')
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [('/l0.csv', 'l2.csv/l0.csv')])

        actual = get_download_src_dst_list('l0.csv', 'l2.csv/local.csv')
        actual = self.norm_src_dst_list(actual)
        self.assertFileListEqual(actual, [('/l0.csv', 'l2.csv/local.csv')])

        with self.assertRaises(SharedFileException) as e:
            actual = get_download_src_dst_list('l0.csv/', '.')
        self.assertEqual(str(e.exception), 'Not a directory: /l0.csv/')

        with self.assertRaises(SharedFileException) as e:
            actual = get_download_src_dst_list('l0.csv', 'not_exist/')
        self.assertEqual(str(e.exception), 'No such file or directory: not_exist')

        with self.assertRaises(SharedFileException) as e:
            actual = get_download_src_dst_list('l0.csv', 'not_exist/local.csv')
        self.assertEqual(str(e.exception), 'No such file or directory: not_exist')

        with self.assertRaises(SharedFileException) as e:
            actual = get_download_src_dst_list('ln.csv', '.')
        self.assertEqual(str(e.exception), 'No such file or directory: /ln.csv')

    def norm_src_dst_list(self, pairs):
        return [(p[0], os.path.normpath(p[1])) for p in pairs]

    def test_normal_src_path(self):
        # normalize the user input source path
        self.assertEqual('/abc', _normalize_user_input_path('./abc'))
        self.assertEqual('/abc', _normalize_user_input_path('abc'))
        self.assertEqual('/', _normalize_user_input_path(''))
        self.assertEqual('/', _normalize_user_input_path('.'))
        self.assertEqual('/', _normalize_user_input_path('./'))
        self.assertEqual('.abc', _normalize_user_input_path('.abc'))

    def test_normal_dest_path(self):
        self.assertEqual('.', _normalize_dest_path(''))
        self.assertEqual('.', _normalize_dest_path('.'))
        self.assertEqual('.', _normalize_dest_path('./'))
        self.assertEqual('/', _normalize_dest_path('/'))
        self.assertEqual('abc', _normalize_dest_path('abc'))
        self.assertEqual('./abc', _normalize_dest_path('./abc'))



from functools import wraps
from inspect import signature
from types import FunctionType
from typing import Dict

from primehub.utils import create_logger
from primehub.utils.optionals import default_optional_builder

logger = create_logger('decorator')

__command_groups__: Dict[str, list] = dict()
__actions__: Dict[str, dict] = dict()
__requires_permission__: Dict[str, str] = dict()


def find_actions(sub_command):
    m = sub_command.__module__
    if m in __command_groups__:
        actions = __command_groups__[m]
        from operator import itemgetter
        return sorted(actions, key=itemgetter('name'))
    return []


def _a(module, action):
    return "{}.{}".format(module.__module__, action)


def find_action_info(module, action):
    logger.debug("find_action_info (%s, %s)", module, action)
    if _a(module, action) in __actions__:
        return __actions__[_a(module, action)]


def find_action_method(module, action):
    info = find_action_info(module, action)
    if info:
        return info['func']


def register_to_command_group(func, cmd_args):
    """
    A python module is a .py file, there are one or more commands in a module.
    A module is a command group with lots of actions and each action is a method in the module.

    For example:
    module cli_config could have {set_endpoint, set_token_, set_group} actions.

    We keep:
     * an action with the key "<modulde>.<action>" for O(1) action search.
     * an module with the key "<module>" for O(1) module-actions search
    """

    if func.__module__ not in __command_groups__:
        __command_groups__[func.__module__] = []

    __command_groups__[func.__module__].append(cmd_args)
    __actions__["{}.{}".format(func.__module__, cmd_args['name'])] = cmd_args


def cmd(**cmd_args):
    if 'name' not in cmd_args:
        raise ValueError('name field is required')
    if 'description' not in cmd_args:
        raise ValueError('name description is required')
    if 'optionals' not in cmd_args:
        cmd_args['optionals'] = []
    if 'return_required' not in cmd_args:
        cmd_args['return_required'] = False

    for idx, opts in enumerate(cmd_args['optionals']):
        maybe_builder = opts[-1]
        if not isinstance(maybe_builder, FunctionType):
            cmd_args['optionals'][idx] = tuple(list(opts) + [default_optional_builder])

    def inner(func):
        make_command_references(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def make_command_references(func):
        # TODO only generate references when invoked from primehub-cli
        cmd_args['module'] = func.__module__
        cmd_args['func'] = func.__name__
        sig = signature(func)
        import inspect

        def info(x):
            if x.annotation is inspect.Parameter.empty:
                t = str
            else:
                t = x.annotation
            return x.name, t, str(x.kind) == 'VAR_KEYWORD',

        cmd_args['arguments'] = [info(x) for x in sig.parameters.values() if x.name != 'self']
        logger.debug("cmd -> %s", cmd_args)
        register_to_command_group(func, cmd_args)

    return inner


def show_debug_info():
    print("Command Groups:")
    for k, v in __command_groups__.items():
        print("  {}".format(k))
        for a in v:
            print("    {}".format(a))
    print("")
    print("Actions:")
    for k, v in __actions__.items():
        print("  {}\n    {}".format(k, v))
    print("")
    print("Permission:")
    for k, v in __requires_permission__.items():
        print("  {}".format(k))
    print("")


class Command(object):

    def __init__(self, name, description, optional_arguments: list):
        self.name = name
        self.description = description
        self.optional_arguments = optional_arguments



import json
from typing import Iterator

from primehub import Helpful, cmd, Module, primehub_load_config
from primehub.utils import resource_not_found, PrimeHubException
from primehub.utils.optionals import file_flag
from primehub.utils.permission import ask_for_permission


def _error_handler(response):
    import re

    if 'errors' in response:
        message = [x for x in response['errors'] if 'message' in x]
        if message:
            message = message[0]['message']
            result = re.findall(r'phschedules.primehub.io "([^"]+)" not found', message)
            if result:
                resource_not_found('schedule', result[0], 'id')


def invalid_config(message: str):
    example = """
    {"instanceType":"cpu-1","image":"base-notebook","displayName":"schedule-example","command":"echo 'good job'","recurrence":{"type":"daily","cron":"0 4 * * *"}}
    """.strip()  # noqa: E501
    raise PrimeHubException(message + "\n\nExample:\n" + json.dumps(json.loads(example), indent=2))


def verify_config(config):
    # verify required fields in the config
    if 'instanceType' not in config:
        invalid_config('instanceType is required')
    if 'image' not in config:
        invalid_config('image is required')
    if 'displayName' not in config:
        invalid_config('displayName is required')
    if 'command' not in config:
        invalid_config('command is required')
    if 'recurrence' not in config:
        invalid_config('recurrence is required')


class Schedules(Helpful, Module):
    """
    The schedules module provides functions to manage PrimeHub Schedules

    Schedule configuration example:
    {
        "instanceType": "cpu-1",
        "image": "base-notebook",
        "displayName": "test",
        "command": "echo \"test\"",
        "recurrence": {
            "type":"daily",
            "cron":"0 4 * * *"
        }
    }
    """

    def _verify_dependency(self, config):
        if 'instanceType' in config:
            self.primehub.instancetypes.get(config['instanceType'])

    @cmd(name='list', description='List schedules', optionals=[('page', int)])
    def list(self, **kwargs) -> Iterator[dict]:
        """
        List all schedules information in the current group

        :type page: int
        :param page: The page number as you can see in PrimeHub Schedules UI

        :rtype list
        :return The list of schedules
        """
        query = """
        query ($where: PhScheduleWhereInput, $page: Int, $orderBy: PhScheduleOrderByInput) {
          phSchedulesConnection(where: $where, page: $page, orderBy: $orderBy) {
            pageInfo {
              totalPage
              currentPage
            }
            edges {
              cursor
              node {
                id
                displayName
                recurrence {
                  type
                  cron
                }
                invalid
                message
                command
                groupId
                groupName
                image
                instanceType {
                  id
                  name
                  displayName
                  cpuLimit
                  memoryLimit
                  gpuLimit
                }
                userId
                userName
                nextRunTime
              }
            }
          }
        }
        """
        variables = {'where': {'groupId_in': [self.group_id]}, 'page': 1}
        page = kwargs.get('page', 0)
        if page:
            variables['page'] = page
            results = self.request(variables, query)
            for e in results['data']['phSchedulesConnection']['edges']:
                yield e['node']
            return

        page = 1
        while True:
            variables['page'] = page
            results = self.request(variables, query)
            if results['data']['phSchedulesConnection']['edges']:
                for e in results['data']['phSchedulesConnection']['edges']:
                    yield e['node']
                page = page + 1
            else:
                break

    @cmd(name='get', description='Get a schedule by id', return_required=True)
    def get(self, id):
        """
        Get detail information of a schedule by id

        :type id: str
        :param id: The schedule id

        :rtype dict
        :return The detail information of a schedule
        """
        query = """
        query ($where: PhScheduleWhereUniqueInput!) {
          phSchedule(where: $where) {
            id
            displayName
            recurrence {
              type
              cron
            }
            invalid
            message
            command
            groupId
            groupName
            image
            instanceType {
              id
              name
              displayName
              cpuLimit
              memoryLimit
              gpuLimit
            }
            userId
            userName
            nextRunTime
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['phSchedule']

    @cmd(name='create', description='Create a schedule', optionals=[('file', file_flag)])
    def _create_cmd(self, **kwargs):
        """
        Submit a schedule from commands

        :type file: str
        :param file: The file path of schedule configurations

        :rtype dict
        :return The detail information of the created schedule
        """
        config = primehub_load_config(filename=kwargs.get('file', None))

        if not config:
            invalid_config('Schedule description is required.')

        return self.create(config)

    def create(self, config):
        """
        Create a schedules with config

        :type config: dict
        :param config: The schedule config

        :rtype dict
        :return The detail information of the created schedule
        """
        query = """
        mutation ($data: PhScheduleCreateInput!) {
          createPhSchedule(data: $data) {
            id
            displayName
            recurrence {
              type
              cron
            }
            invalid
            message
            command
            groupId
            groupName
            image
            instanceType {
              id
              name
              displayName
              cpuLimit
              memoryLimit
              gpuLimit
            }
            userId
            userName
            nextRunTime
          }
        }
        """
        config['groupId'] = self.group_id

        verify_config(config)
        self._verify_dependency(config)
        results = self.request({'data': config}, query)
        return results['data']['createPhSchedule']

    @cmd(name='update', description='Update a schedule by id', optionals=[('file', file_flag)])
    def _update_cmd(self, id, **kwargs):
        """
        Update a schedule from commands

        :type file: str
        :param file: The file path of schedule configurations

        :rtype dict
        :return The detail information of the updated schedule
        """
        config = primehub_load_config(filename=kwargs.get('file', None))
        if not config:
            invalid_config('Schedule description is required.')

        return self.update(id, config)

    def update(self, id, config):
        """
        Update a schedule with config

        :type id: str
        :param id: The schedule id

        :type config: dict
        :param config: The schedule config

        :rtype dict
        :return The detail information of the updated schedule
        """
        query = """
        mutation ($data: PhScheduleUpdateInput!, $where: PhScheduleWhereUniqueInput!) {
          updatePhSchedule(data: $data, where: $where) {
            id
            displayName
            recurrence {
              type
              cron
            }
            invalid
            message
            command
            groupId
            groupName
            image
            instanceType {
              id
              name
              displayName
              cpuLimit
              memoryLimit
              gpuLimit
            }
            userId
            userName
            nextRunTime
          }
        }
        """
        config['groupId'] = self.group_id
        if not config:
            invalid_config('Schedule description is required.')
        self._verify_dependency(config)
        results = self.request({'data': config, 'where': {'id': id}}, query, _error_handler)
        return results['data']['updatePhSchedule']

    @ask_for_permission
    @cmd(name='delete', description='Delete a schedule by id')
    def delete(self, id, **kwargs):
        """
        Delete a schedule by id

        :type id: str
        :param id: The schedule id

        :rtype dict
        :return The detail information of the deleted schedule
        """
        query = """
        mutation ($where: PhScheduleWhereUniqueInput!) {
          deletePhSchedule(where: $where) {
            id
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['deletePhSchedule']

    def help_description(self):
        return "Get a schedule or list schedules"



import abc
import importlib
import json
import os
import sys
from typing import Union, Callable, Dict, Any

from primehub.utils import group_required, create_logger
from primehub.utils.decorators import cmd  # noqa: F401
from primehub.utils.display import Display, HumanFriendlyDisplay, Displayable
from primehub.utils.http_client import Client

logger = create_logger('primehub-config')


def _get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'VERSION'))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


__version__ = _get_version()


class PrimeHubConfig(object):
    """
    PrimeHubConfig load the config from the default path ~/.primehub/config.json

    The config.json looks like:
    {
       "endpoint": ""
       "api-token": "",
       "group": {
          "id": "",
          "name": "",
          "displayName": "",
       }
    }

    PrimeHubConfig allows changing setting from four ways:
    * the default config path
    * alternative path for the config file (config argument from constructor)
    * environment variables: PRIMEHUB_API_TOKEN, PRIMEHUB_API_ENDPOINT and PRIMEHUB_GROUP
    * set property for api_token, endpoint and group

    PrimeHubConfig evaluates a property in the above order and the last updates take effect
    """

    def __init__(self, **kwargs):
        self.config_file = kwargs.get('config', None)
        if not self.config_file:
            self.config_file = self.get_default_path()

        # PrimeHub SDK evaluate
        self.config_from_file = {}
        self.config_from_env = {}
        self.config_from_user_input = {}
        self.group_info = {}

        self.load_config()
        self.load_config_from_env()
        self.set_properties(**kwargs)

    def set_properties(self, **kwargs):
        if kwargs.get('group', None):
            self.group = kwargs['group']
        if kwargs.get('token', None):
            self.api_token = kwargs['token']
        if kwargs.get('endpoint', None):
            self.endpoint = kwargs['endpoint']

    def load_config(self):
        try:
            if not os.path.exists(os.path.expanduser(self.config_file)):
                return
            with open(self.config_file, "r") as fh:
                self.config_from_file = json.load(fh)
                if self.config_from_file and 'group' in self.config_from_file:
                    self.group_info = self.config_from_file['group']
        except BaseException:
            pass

    def load_config_from_env(self):
        # environment variables: PRIMEHUB_API_TOKEN, PRIMEHUB_API_ENDPOINT and PRIMEHUB_GROUP

        def set_env(key):
            if os.environ.get(key):
                self.config_from_env[key] = os.environ.get(key)
                if key == 'PRIMEHUB_GROUP' and self.group_info:
                    self.group_info['name'] = os.environ.get(key)

        set_env('PRIMEHUB_API_TOKEN')
        set_env('PRIMEHUB_API_ENDPOINT')
        set_env('PRIMEHUB_GROUP')

    def get_default_path(self):
        return os.path.expanduser("~/.primehub/config.json")

    def save(self, path=None):
        """
        The config.json looks like:
        {
           "endpoint": ""
           "api-token": "",
           "group": {
              "id": "",
              "name": "",
              "displayName": "",
           }
        }
        """
        output = dict()
        output['endpoint'] = self.endpoint
        output['api-token'] = self.api_token
        if self.group_info and self.group_info.get('name', None) == self.group:
            output['group'] = self.group_info
        else:
            output['group'] = dict(name=self.group)

        output_path = os.path.expanduser(path or self.config_file)
        if os.path.dirname(output_path):
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, "w") as fh:
            fh.write(json.dumps(output, indent=2, sort_keys=True))

    @property
    def group(self):
        if self.config_from_user_input.get('group', None):
            logger.debug('group config_from_user_input')
            return self.config_from_user_input['group']
        if self.config_from_env.get('PRIMEHUB_GROUP', None):
            logger.debug('group config_from_env')
            return self.config_from_env.get('PRIMEHUB_GROUP')
        if self.config_from_file.get('group', None) and self.config_from_file['group'].get('name', None):
            logger.debug('group config_from_file')
            return self.config_from_file['group']['name']

    @group.setter
    def group(self, group):
        if group:
            self.config_from_user_input['group'] = group
            if self.group_info:
                self.group_info['name'] = group

    @property
    def api_token(self):
        if self.config_from_user_input.get('api-token', None):
            return self.config_from_user_input['api-token']
        if self.config_from_env.get('PRIMEHUB_API_TOKEN', None):
            return self.config_from_env.get('PRIMEHUB_API_TOKEN')
        return self.config_from_file.get('api-token', None)

    @api_token.setter
    def api_token(self, api_token):
        if api_token:
            self.config_from_user_input['api-token'] = api_token

    @property
    def endpoint(self):
        if self.config_from_user_input.get('endpoint', None):
            return self.config_from_user_input['endpoint']
        if self.config_from_env.get('PRIMEHUB_API_ENDPOINT', None):
            return self.config_from_env.get('PRIMEHUB_API_ENDPOINT')
        return self.config_from_file.get('endpoint', None)

    @endpoint.setter
    def endpoint(self, endpoint):
        if endpoint:
            self.config_from_user_input['endpoint'] = endpoint

    @property
    def current_group(self) -> dict:
        return self.group_info

    @current_group.setter
    def current_group(self, group_info):
        if group_info and group_info.get('id', None):
            self.group_info = group_info


class Helpful(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def help_description(self):
        """one line description for all commands"""
        return NotImplemented


class PrimeHub(object):

    def __init__(self, config: PrimeHubConfig):
        self.primehub_config = config
        self.json_output = True
        self.commands: Dict[str, Module] = dict()
        self._stderr = sys.stderr
        self._stdout = sys.stdout

        # register commands
        self.register_command('config', 'Config')
        self.register_command('groups', 'Groups')
        self.register_command('images', 'Images')
        self.register_command('datasets', 'Datasets')
        self.register_command('instancetypes', 'InstanceTypes')
        self.register_command('jobs', 'Jobs')
        self.register_command('schedules', 'Schedules')
        self.register_command('deployments', 'Deployments')
        self.register_command('notebooks', 'Notebooks')
        self.register_command('files', 'Files')
        self.register_command('me', 'Me')
        self.register_command('version', 'Version')

        # initial
        self._ensure_config_details(config)

    def _ensure_config_details(self, config: PrimeHubConfig):
        try:
            group_id = config.current_group.get('id', None)
            group_name = config.current_group.get('name', None)
            if group_name is None:
                return

            if group_id is None:
                self.config.reconfigure_group(group_name)
        except BaseException:
            pass

    def request(self, variables: dict, query: str, error_handler: Callable = None):
        return Client(self.primehub_config).request(variables, query, error_handler)

    def request_logs(self, endpint: str, follow: bool, tail: int):
        return Client(self.primehub_config).request_logs(endpint, follow, tail)

    def request_file(self, endpint: str, dest: str):
        return Client(self.primehub_config).request_file(endpint, dest)

    def register_command(self, module_name: str, command_class: Union[str, Callable], command_name=None):
        if not command_name:
            command_name = module_name

        # create command instance
        if isinstance(command_class, str):
            clazz = importlib.import_module('primehub.' + module_name).__getattribute__(command_class)
        else:
            clazz = command_class

        # register to the commands table
        self.commands[command_name] = clazz(self)

    def __getattr__(self, item):
        if item in self.commands:
            return self.commands[item]
        raise AttributeError("Cannot find a command [{}]".format(item))

    @property
    def stderr(self):
        return self._stderr

    @stderr.setter
    def stderr(self, out):
        self._stderr = out

    @property
    def stdout(self):
        return self._stdout

    @stdout.setter
    def stdout(self, out):
        self._stdout = out

    def get_all_commands(self):
        return sorted(self.commands.keys())

    def is_ready(self):
        if self.primehub_config.current_group is None:
            return False
        if not self.primehub_config.current_group.get('id', None):
            return False
        return True


class Module(object):

    def __init__(self, primehub: PrimeHub, **kwargs):
        self.primehub = primehub

        # attach request method
        self.request = primehub.request
        self.request_logs = primehub.request_logs
        self.request_file = primehub.request_file

    @property
    def current_group(self) -> dict:
        g = self.primehub.primehub_config.current_group
        if not g:
            group_required()
        if not g.get('id', None):
            group_required()
        return g

    @property
    def group_id(self) -> str:
        return self.current_group['id']

    @property
    def group_name(self) -> str:
        return self.current_group['name']

    @property
    def endpoint(self) -> str:
        return self.primehub.primehub_config.endpoint

    @property
    def primehub_config(self):
        raise ValueError(
            'The attribute [primehub_config] is access denied, '
            'please use props of the Module to get configurations')

    def get_display(self) -> Displayable:
        if self.primehub.json_output:
            return Display()
        else:
            return HumanFriendlyDisplay()

    def display(self, action: dict, value: Any):
        self.get_display().display(action, value, self.primehub.stdout)


def has_data_from_stdin():
    """
    Check if any data comes from stdin.

    :return: True if there are data from stdin, otherwise False
    """
    import sys
    import select

    if select.select([sys.stdin, ], [], [], 0.0)[0]:
        return True
    else:
        return False


def primehub_load_config(filename):
    if has_data_from_stdin():
        return json.loads("".join(sys.stdin.readlines()))

    if filename and os.path.exists(filename):
        with open(filename, 'r') as fh:
            return json.load(fh)

    return {}



import json
import time
from tempfile import mkstemp

from primehub import Helpful, cmd, Module
from primehub.config import Config
from primehub.groups import Groups
from primehub.utils import create_logger

logger = create_logger('e2e')


class E2EForBasicFunction(Helpful, Module):

    def title(self, name):
        print("{} {:15s} {}".format('=' * 30, name, '=' * 30))

    def endline(self):
        print()
        print()

    @cmd(name='basic-functions', description='run basic functions')
    def basic_functions(self):
        self.title('Config')
        c: Config = self.primehub.config
        c.primehub.primehub_config.save(path='./e2e-config.json')
        print('save config to ', './e2e-config.json')
        self.endline()

        self.title('Info')
        print(self.primehub.info.info())
        self.endline()

        self.title('Groups')
        group: Groups = self.primehub.groups
        for g in group.list():
            print(g['id'], g['name'])
        self.endline()

        self.title('Jobs')
        self.job_e2e()

    def job_e2e(self):
        instance_types = [x for x in self.primehub.instancetypes.list() if "gpu" not in x]
        instance_type = None
        for x in instance_types:
            if 'cpu-half' in x['name']:
                instance_type = x['id']
                break
            if 'cpu-1' in x['name']:
                instance_type = x['id']
                break
        print('instance-type:', instance_type)
        scripts = r"""
        sudo apt-get update -y
        sudo apt-get install -y git
        pip install git+https://github.com/InfuseAI/primehub-python-sdk.git@main
        primehub -h
        echo
        exit 0
        """
        job_spec = dict(instanceType=instance_type, image="base-notebook", displayName="job-e2e", command=scripts)
        fd, path = mkstemp(".json")
        with open(path, "w") as fh:
            fh.write(json.dumps(job_spec))

        my_job = self.primehub.jobs._submit_cmd(file=path)
        my_id = my_job['id']
        print("Job ID:", my_job['id'])

        last_state = None
        while True:
            p = self.primehub.jobs.get(my_id)['phase']
            if last_state is None:
                last_state = p

            if p != last_state:
                print("Job Phase: {} -> {}".format(last_state, p))
                last_state = p

            if last_state == 'Running':
                break

            if last_state == 'Succeeded':
                break

            if last_state == 'Cancelled':
                break
            time.sleep(1)

        print("Logs:")
        for g in self.primehub.jobs.logs(my_id, follow=True):
            print(g)

    def help_description(self):
        return "dev-lab is used to the primehub-python-sdk development and testing"



import json

from primehub import Helpful, Module, cmd, PrimeHubConfig
from tests import BaseTestCase


class ConfigWatcher(Helpful, Module):

    @cmd(name='watch', description='watch-primehub-config')
    def watch(self):
        # ensure the config file is changed
        assert self.primehub.primehub_config.config_file != self.primehub.primehub_config.get_default_path()

        from tempfile import mkstemp
        fd, path = mkstemp('.json', text=True)
        self.primehub.primehub_config.save(path=path)

        return dict(config_file=path)

    def help_description(self):
        return "Big Brother Watch YOU"


class TestCliConfig(BaseTestCase):

    def setUp(self) -> None:
        super(TestCliConfig, self).setUp()

        # clean commands, add the FakeCommand
        self.sdk.register_command('test_cli_config', ConfigWatcher)

    def test_config_options(self):
        """
        Global Options:
          --config CONFIG      the path of the config file
          --endpoint ENDPOINT  the endpoint to the PrimeHub GraphQL URL
          --token TOKEN        API Token generated from PrimeHub Console
          --group GROUP        override the active group
        """
        group_info = {'name': 'test-config-from-cli:group'}
        self.mock_request.return_value = {'data': {'me': {'effectiveGroups': [group_info]}}}
        c = self.make_cfg()

        # Verify the saved config will have test-config-from-cli:*
        self.assert_config_with_prefix('test-config-from-cli', c)

        # Verify other options
        group_info = {'name': 'opt:group'}
        self.mock_request.return_value = {'data': {'me': {'effectiveGroups': [group_info]}}}
        c = self.make_cfg(['--endpoint', 'opt:endpoint', '--token', 'opt:api-token', '--group', 'opt:group'])
        self.assert_config_with_prefix('opt', c)

    def make_cfg(self, extra_args: list = None):
        if extra_args is None:
            extra_args = []

        cfg_path = self.create_fake_config(self.cfg_dict_with_prefix('test-config-from-cli'))
        output = self.cli_stdout(['app.py', 'test_cli_config', 'watch', '--config', cfg_path] + extra_args)

        # bug: the output buffer contains the previous result, we only take the last one
        output = output.strip().split("\n")[-1]

        saved_cfg_path = json.loads(output)['config_file']
        new_cfg = PrimeHubConfig(config=saved_cfg_path)
        return new_cfg



from typing import Optional

from primehub import Helpful, cmd, Module
from primehub.resource_operations import GroupResourceOperation


class Images(Helpful, Module, GroupResourceOperation):
    """
    List images or get an image from the list
    """
    resource_name = 'images'
    query = """
    {
      me {
        effectiveGroups {
          name
          images {
            id
            name
            displayName
            description
            useImagePullSecret
            spec
          }
        }
      }
    }
    """

    @cmd(name='list', description='List images')
    def list(self) -> list:
        """
        List images

        :rtype: list
        :returns: all images in the current group
        """
        return self.do_list(Images.query, Images.resource_name)

    @cmd(name='get', description='Get a image by name', return_required=True)
    def get(self, name) -> Optional[dict]:
        """
        Get an image from the current group

        :type name: str
        :param name: the name of an image

        :rtype: Optional[dict]
        :returns: an image
        """
        return self.do_get(Images.query, Images.resource_name, name)

    def help_description(self):
        return "Get a image or list images"



from primehub import PrimeHubConfig
from tests import BaseTestCase


class TestCmdConfig(BaseTestCase):

    def setUp(self) -> None:
        super(TestCmdConfig, self).setUp()

    def test_config_options(self):
        """
        Usage:
          primehub config [command]

        Update the settings of PrimeHub SDK

        Available Commands:
          set-endpoint         set endpoint and save to the config file
          set-token            set token and save to the config file
          set-group            set group and save to the config file
        """

        # use cli version's config
        self.sdk.register_command('config', 'CliConfig')

        args = ['app.py', 'config', 'set-endpoint', 'config-set:endpoint']
        self.cli_stdout(args)

        args = ['app.py', 'config', 'set-token', 'config-set:api-token']
        self.cli_stdout(args)

        # mock the result of fetching group info
        group_info = {
            'id': 'group-id',
            'name': 'config-set:group',
            'displayName': 'group-display-name'
        }
        self.mock_request.return_value = {'data': {'me': {'effectiveGroups': [group_info]}}}
        args = ['app.py', 'config', 'set-group', 'config-set:group']
        self.cli_stdout(args)

        # Verify new configuration has been prefixed with 'config:set'
        self.assert_config_with_prefix('config-set', PrimeHubConfig())

        # Verify group information updated
        # {
        #   me {
        #     effectiveGroups {
        #       id
        #       name
        #       displayName
        #     }
        #   }
        # }
        self.assertEqual(group_info, self.load_json_file(PrimeHubConfig().get_default_path())['group'])



from argparse import ArgumentParser


def default_optional_builder(parser: ArgumentParser, name: str, type_of_arg: type):
    parser.add_argument("--" + name, type=type_of_arg)


def toggle_flag(parser: ArgumentParser, name: str):
    parser.add_argument("--" + name, action="store_true", default=False)


def file_flag(parser: ArgumentParser, name: str):
    if name != 'file':
        raise ValueError(f'name should be file, but it is {name}')
    parser.add_argument("--file", "-f", dest='file')



from primehub import Helpful, cmd, Module, __version__


class Version(Helpful, Module):

    @cmd(name='version', description='show version number', return_required=True)
    def version(self) -> str:
        return __version__

    def help_description(self):
        return "Display the version of PrimeHub Python SDK"



from primehub import Helpful, cmd, Module
from urllib.parse import urlparse
import os

from primehub.utils.optionals import toggle_flag
from primehub.utils import create_logger, SharedFileException

logger = create_logger('cmd-files')


def invalid(message):
    raise SharedFileException(message)


def _normalize_dest_path(path):
    if path is None:
        raise ValueError('path is required')

    # case empty string => .
    if path == '':
        return '.'

    # simple normalized the typo to .
    if path in ['.', './']:
        return '.'

    # the normal case
    if path == '/':
        return '/'

    # case ./abc => /abc
    if path.startswith('./'):
        return path

    # case .abc => .abc
    if path.startswith('.'):
        return path

    return path


def _normalize_user_input_path(path):
    if path is None:
        raise ValueError('path is required')

    # case empty string => /
    if path == '':
        return '/'

    # simple normalized the typo to /
    if path in ['.', './']:
        return '/'

    # the normal case
    if path == '/':
        return '/'

    # case ./abc => /abc
    if path.startswith('./'):
        return '/' + path[2:]

    # case .abc => .abc
    if path.startswith('.'):
        return path

    # case abc => /abc
    if not path.startswith('/'):
        return '/' + path

    return path


class Files(Helpful, Module):
    """
    The files module provides functions to manage Primehub Shared Files
    """

    @cmd(name='list', description='List shared files')
    def list(self, path):
        """
        The cmd to list all files and folders in the path

        :type path: str
        :param path: The path to list

        :rtype dict
        :return The detail information of files in the path
        """

        items = self._execute_list(path, limit=1)
        if items:  # directory
            return self._execute_list(path)

        items = self._execute_list(path, recursive=True, limit=1)
        if not items or items[0]['name']:
            invalid(f'No such file or directory: {path}')
            return []

        # file
        if not os.path.basename(path):  # trailing slash
            invalid(f'Not a directory: {path}')
            return []

        items[0]['name'] = os.path.basename(path)
        return items

    def _execute_list(self, path, **kwargs):
        """
        List all files and folders in the path

        :type path: str
        :param path: The path to list

        :type recursive: bool
        :param recursive: List recursively, it works when a path is a directory.

        :type limit: int
        :param limit: The maximum size of the list

        :rtype dict
        :return The detail information of files in the path
        """
        query = """
        query files($where: StoreFileWhereInput!, , $options: StoreFileListOptionInput) {
          files (where: $where, options: $options) {
            phfsPrefix
            items {
              name
              size
              lastModified
            }
          }
        }
        """
        path = _normalize_user_input_path(path)

        path_norm = os.path.normpath(path)
        recursive = kwargs.get('recursive', False)
        limit = kwargs.get('limit', 1000)
        results = self.request(
            {'where': {'phfsPrefix': path_norm, 'groupName': self.group_name},
             'options': {'recursive': recursive, 'limit': limit}},
            query)
        items = results['data']['files']['items']
        return items

    @cmd(name='download', description='Download shared files', optionals=[('recursive', toggle_flag)])
    def download(self, path, dest, **kwargs):
        """
        Download files

        :type path: str
        :param path: The path of file or folder

        :type dest: str
        :param dest: The local path to save artifacts

        :type recusive: bool
        :param recusive: Copy recursively, it works when a path is a directory.
        """

        def to_group_path(group_name: str):
            if not group_name:
                return group_name
            return group_name.replace('_', '-').lower()

        u = urlparse(self.endpoint)
        endpoint = u._replace(path=f'/api/files/groups/{to_group_path(self.group_name)}').geturl()

        # start download
        src_dst_list = self._generate_download_list(path, dest, **kwargs)
        for src, dst in src_dst_list:
            dir = os.path.dirname(dst)
            if dir and not os.path.isdir(dir):
                os.makedirs(dir)
            self.request_file(endpoint + src, dst)

    def _generate_download_list(self, path, dest, **kwargs):
        """
        Download files

        :type path: str
        :param path: The path of file or folder

        :type dest: str
        :param dest: The local path to save artifacts

        :type recusive: bool
        :param recusive: Copy recursively, it works when a path is a directory.

        :type list
        :return List of tuple of download source and destination
        """
        path = _normalize_user_input_path(path)
        recursive = kwargs.get('recursive', False)

        # check dest
        dest = _normalize_dest_path(dest)
        dest_norm = os.path.normpath(dest)
        dest_isfile = os.path.isfile(dest_norm)
        dest_dir = os.path.dirname(dest)
        if dest_dir and not os.path.isdir(dest_dir):
            invalid(f'No such file or directory: {dest_dir}')
            return []

        items = self._execute_list(path, limit=1)
        if items:  # directory
            if dest_isfile:
                invalid(f'Not a directory: {dest}')
                return []

            if not recursive:
                invalid(f'{path} is a directory, please download it recursively')
                return []

            transform = not any(os.path.basename(path))

        else:  # file or not exist
            items = self._execute_list(path, recursive=True, limit=1)
            if not items or items[0]['name']:
                invalid(f'No such file or directory: {path}')
                return []

            if not os.path.basename(path):  # trailing slash
                invalid(f'Not a directory: {path}')
                return []

            transform = not os.path.isdir(dest)

        src_dst_list = []
        path_norm = os.path.normpath(path)
        prefix = path_norm if transform else os.path.dirname(path_norm)
        prefix_len = len(os.path.join(prefix, ''))

        files_phfs = [path_norm + f['name'] for f in self._execute_list(path_norm, recursive=True)]
        for src in files_phfs:
            dst = os.path.normpath(os.path.join(dest_norm, src[prefix_len:]))
            if os.path.isdir(dst):
                logger.warning(f'cannot overwrite directory {dst} with non-directory {src}')
                continue

            is_file = False
            sub_dst = dest_norm
            dirs = src[prefix_len:].split('/')
            for dir in dirs[:-1]:
                sub_dst = os.path.join(sub_dst, dir)
                if os.path.isfile(sub_dst):
                    is_file = True
                    break
                if not os.path.exists(sub_dst):
                    break
            if is_file:
                logger.warning(f'{dest} Not a directory')
                continue

            src_dst_list.append((src, dst))

        return src_dst_list

    def help_description(self):
        return "List and download shared files"



import json
import time
from typing import Iterator

from primehub import Helpful, cmd, Module, primehub_load_config
from primehub.utils import resource_not_found, PrimeHubException
from primehub.utils.optionals import toggle_flag, file_flag
from primehub.utils.permission import ask_for_permission


def _error_handler(response):
    import re

    if 'errors' in response:
        message = [x for x in response['errors'] if 'message' in x]
        if message:
            message = message[0]['message']
            result = re.findall(r'phdeployments.primehub.io "([^"]+)" not found', message)
            if result:
                resource_not_found('schedule', result[0], 'id')


def invalid_config(message: str):
    import random
    import string
    import textwrap

    id = 'quickstart-iris-' + ''.join(random.choices(string.ascii_lowercase + string.digits, k=5))
    example = """
    {"name":"quickstart-iris","modelImage":"infuseai/sklearn-prepackaged:v0.1.0","modelURI":"gs://seldon-models/sklearn/iris","env":[],"metadata":{},"instanceType":"cpu-1","replicas":1,"updateMessage":"","id":"<id>","endpointAccessType":"public"}
    """.strip()  # noqa: E501
    example = example.replace('<id>', id)

    docs = """
    We take examples from:
    https://docs.primehub.io/docs/model-deployment-tutorial-prepackaged-image

    Definition example:
    """
    docs = textwrap.dedent(docs)
    explain = f'{message}\n\n{docs}{json.dumps(json.loads(example), indent=2)}'
    raise PrimeHubException(explain)


def verify_requires(config):
    # verify required fields in the config
    if not config:
        invalid_config('Deployment definition is required.')
    if 'id' not in config:
        invalid_config('id is required')
    if 'instanceType' not in config:
        invalid_config('instanceType is required')
    if 'modelImage' not in config:
        invalid_config('modelImage is required')


class Deployments(Helpful, Module):
    """
    The deployments module provides functions to manage PrimeHub Deployments

    Deployment configuration example:
    {
        "id": "<uniqe_id>",
        "name": "deploy-name",
        "modelImage": "base-notebook",
        "modelURI": "test/module/uri",
        "instanceType": "cpu-1",
        "replicas": 1
    }
    """

    @cmd(name='list', description='List deployments')
    def list(self):
        """
        List all deployments information in the current group

        :rtype list
        :return The list of deployments
        """
        query = """
        query($where: PhDeploymentWhereInput) {
          phDeployments(where: $where) {
            id
            name
            modelImage
            imagePullSecret
            description
            replicas
            stop
            endpointAccessType
            endpointClients {
              name
            }
            status
            endpoint
            replicas
            availableReplicas
            message
            pods {
              name
            }
          }
        }
        """
        results = self.request({'where': {'groupId_in': [self.group_id]}}, query)
        return results['data']['phDeployments']

    @cmd(name='get', description='Get a deployment by id', return_required=True)
    def get(self, id):
        """
        Get detail information of a deployment by id

        :type id: str
        :param id: The deployment id

        :rtype dict
        :return The detail information of a deployment
        """
        query = """
        query ($where: PhDeploymentWhereUniqueInput!) {
          phDeployment (where: $where) {
            id
            name
            modelImage
            imagePullSecret
            description
            replicas
            stop
            endpointAccessType
            endpointClients {
              name
            }
            status
            endpoint
            replicas
            availableReplicas
            message
            pods {
              name
            }
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['phDeployment']

    @cmd(name='get-history', description='Get history of a deployment by id')
    def get_history(self, id):
        """
        Get history of a deployment by id

        :type id: str
        :param id: The deployment id

        :rtype dict
        :return The history of a deployment
        """
        query = """
        query ($where: PhDeploymentWhereUniqueInput!) {
          phDeployment (where: $where) {
            history {
              time
              deployment {
                id
                name
                modelImage
                imagePullSecret
                description
                replicas
                stop
                endpointAccessType
                endpointClients {
                  name
                }
                status
                endpoint
                replicas
              }
            }
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['phDeployment']['history']

    @cmd(name='create', description='Create a deployment', optionals=[('file', file_flag)])
    def _create_cmd(self, **kwargs):
        """
        Create a deployment from commands

        :type file: str
        :param file: The file path of deployment configurations

        :rtype dict
        :return The detail information of the created deployment
        """
        config = primehub_load_config(filename=kwargs.get('file', None))
        return self.create(config)

    def create(self, config):
        """
        Create a deployments with config

        :type config: dict
        :param config: The deployment config

        :rtype dict
        :return The detail information of the created deployment
        """
        query = """
        mutation ($data: PhDeploymentCreateInput!) {
          createPhDeployment(data: $data) {
            id
            name
            modelImage
            imagePullSecret
            description
            replicas
            stop
            endpointAccessType
            endpointClients {
              name
            }
            status
            endpoint
            replicas
            availableReplicas
            message
            pods {
              name
            }
          }
        }
        """

        verify_requires(config)
        config['groupId'] = self.group_id
        self.verify_dependency(config)
        results = self.request({'data': config}, query)
        return results['data']['createPhDeployment']

    def verify_dependency(self, config):
        if 'instanceType' in config:
            self.primehub.instancetypes.get(config['instanceType'])

    @cmd(name='update', description='Update a deployment by id', optionals=[('file', file_flag)])
    def _update_cmd(self, id, **kwargs):
        """
        Update a deployment from commands

        :type file: str
        :param file: The file path of deployment configurations

        :rtype dict
        :return The detail information of the updated deployment
        """
        config = primehub_load_config(filename=kwargs.get('file', None))
        return self.update(id, config)

    def update(self, id, config):
        """
        Update a deployment with config

        :type id: str
        :param id: The deployment id

        :type config: dict
        :param config: The deployment config

        :rtype dict
        :return The detail information of the updated deployment
        """
        query = """
        mutation ($data: PhDeploymentUpdateInput!, $where: PhDeploymentWhereUniqueInput!) {
          updatePhDeployment(data: $data, where: $where) {
            id
            name
            modelImage
            imagePullSecret
            description
            replicas
            stop
            endpointAccessType
            endpointClients {
              name
            }
            status
            endpoint
            replicas
            availableReplicas
            message
            pods {
              name
            }
          }
        }
        """
        if not config:
            invalid_config('Deployment definition is required.')
        self.verify_dependency(config)
        results = self.request({'data': config, 'where': {'id': id}}, query, _error_handler)
        return results['data']['updatePhDeployment']

    @cmd(name='start', description='Start a deployment by id')
    def start(self, id):
        """
        Start a deployment by id

        :type id: str
        :param id: The deployment id

        :rtype dict
        :return The detail information of the started deployment
        """
        query = """
        mutation ($where:PhDeploymentWhereUniqueInput!) {
          deployPhDeployment (where: $where) {
            id
            name
            modelImage
            imagePullSecret
            description
            replicas
            stop
            endpointAccessType
            endpointClients {
              name
            }
            status
            endpoint
            replicas
            availableReplicas
            message
            pods {
              name
            }
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['deployPhDeployment']

    @cmd(name='stop', description='Stop a deployment by id')
    def stop(self, id):
        """
        Stop a deployment by id

        :type id: str
        :param id: The deployment id

        :rtype dict
        :return The detail information of the stopped deployment
        """
        query = """
        mutation ($where:PhDeploymentWhereUniqueInput!) {
          stopPhDeployment (where: $where) {
            id
            name
            modelImage
            imagePullSecret
            description
            replicas
            stop
            endpointAccessType
            endpointClients {
              name
            }
            status
            endpoint
            replicas
            availableReplicas
            message
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['stopPhDeployment']

    @ask_for_permission
    @cmd(name='delete', description='Delete a deployment by id')
    def delete(self, id, **kwargs):
        """
        Delete a deployment by id

        :type id: str
        :param id: The deployment id

        :rtype dict
        :return The detail information of the deleted deployment
        """
        query = """
        mutation ($where: PhDeploymentWhereUniqueInput!) {
          deletePhDeployment(where: $where) {
            id
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['deletePhDeployment']

    # TODO: handle invalid pod
    @cmd(name='logs', description='Get deployment logs by id',
         optionals=[('pod', str), ('follow', toggle_flag), ('tail', int)])
    def logs(self, id, **kwargs) -> Iterator[bytes]:
        """
        Get logs of a deployment

        :type id: str
        :param id: The job id

        :type pod: str
        :param pod: The target pod to log

        :type follow: bool
        :param follow: Wait for additional logs to be appended

        :type tail: int
        :param tail: Show last n lines

        :rtype str
        :return The deployment logs
        """
        query = """
        query ($where: PhDeploymentWhereUniqueInput!) {
          phDeployment (where: $where) {
            pods {
              name
              phase
              logEndpoint
            }
          }
        }
        """

        pod_name = kwargs.get('pod', 'deploy-' + id)
        follow = kwargs.get('follow', False)
        tail = kwargs.get('tail', 10)

        results = self.primehub.request({'where': {'id': id}}, query, _error_handler)
        pods = results['data']['phDeployment']['pods']
        endpoints = [p['logEndpoint'] for p in pods if p['name'].startswith(pod_name)]
        return self.primehub.request_logs(endpoints[0], follow, tail)

    @cmd(name='wait', description='Wait a deployment to complete', optionals=[('timeout', int)])
    def wait(self, id, **kwargs):
        """
        Wait a deployment in a stable state {Deployed, Stopped} or until timeout

        :type id: str
        :param id: The job id

        :type timeout: int
        :param timeout: The timeout in second

        :rtype dict
        :return The detail information of the deployment
        """
        query = """
        query ($where: PhDeploymentWhereUniqueInput!) {
          phDeployment (where: $where) {
            stop
            status
          }
        }
        """
        timeout = kwargs.get('timeout', 0)
        start_time = time.time()
        while True:
            results = self.request({'where': {'id': id}}, query, _error_handler)
            status = results['data']['phDeployment']['status']
            stop = results['data']['phDeployment']['stop']
            if (not stop and status == 'Deployed') or (stop and status == 'Stopped'):
                break
            time.sleep(1)
            if timeout != 0 and (time.time() - start_time >= timeout):
                break
        return self.get(id)

    def help_description(self):
        return "Get a deployment or list deployments"



import logging
import os
import sys


class PrimeHubException(BaseException):
    pass


class GraphQLException(PrimeHubException):
    pass


class GroupIsRequiredException(PrimeHubException):
    pass


class UserRejectAction(PrimeHubException):
    pass


class RequestException(PrimeHubException):
    pass


class ResponseException(PrimeHubException):
    pass


class PrimeHubReturnsRequiredException(PrimeHubException):
    pass


class ResourceNotFoundException(PrimeHubException):
    pass


class SharedFileException(PrimeHubException):
    pass


def reject_action(action):
    raise UserRejectAction(
        'User rejects action [%s], please use the flag "--yes-i-really-mean-it" to allow the action.' % action)


def group_required():
    raise GroupIsRequiredException('No group information, please configure the active group first.')


def group_not_found(group):
    raise GroupIsRequiredException('No group information for [%s], please check the configuration again.' % group)


def resource_not_found(resource_type: str, key: str, key_type: str):
    raise ResourceNotFoundException(resource_type, key, key_type)


def create_logger(name) -> logging.Logger:
    log_level = logging.WARNING
    if os.environ.get('PRIMEHUB_SDK_LOG_LEVEL') == 'DEBUG':
        log_level = logging.DEBUG
    if os.environ.get('PRIMEHUB_SDK_LOG_LEVEL') == 'INFO':
        log_level = logging.INFO

    log = logging.getLogger(name)
    log.setLevel(log_level)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(log_level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    log.addHandler(handler)
    return log



import json

from primehub import Module, cmd, Helpful
from primehub.resource_operations import GroupResourceOperation
from tests import BaseTestCase


class GroupFoobarCommand(Helpful, Module, GroupResourceOperation):

    def help_description(self):
        return "foo bar"

    @cmd(name='list', description='')
    def list(self):
        return self.do_list('fake-query', 'foobar')

    @cmd(name='get', description='')
    def get(self, name):
        return self.do_get('fake-query', 'foobar', name)


class TestGroupResourceOperation(BaseTestCase):

    def setUp(self) -> None:
        super(TestGroupResourceOperation, self).setUp()
        self.sdk.register_command('test_group_resource_operation', GroupFoobarCommand)

    def test_group_not_found(self):
        output = self.cli_stderr(['app.py', 'test_group_resource_operation', 'list'])
        self.assertEqual('No group information, please configure the active group first.', output.strip())

    def test_group_resource_operation(self):
        self.sdk.primehub_config.group_info = {'name': 'phusers', 'id': 'any-id'}
        self.mock_request.return_value = {
            'data': {
                'me': {
                    'effectiveGroups': [
                        {'name': 'group-1', 'foobar': [dict(name="1", value="1")]},
                        {'name': 'group-2', 'foobar': [dict(name="2", value="2")]},
                        {'name': 'phusers', 'foobar': [dict(name="3", value="3")]},
                    ]
                }
            }
        }

        # verify list will select the right group
        output = self.cli_stdout(['app.py', 'test_group_resource_operation', 'list'])
        self.assertEqual([dict(name="3", value="3")], json.loads(output))

        # verify get by name
        output = self.cli_stdout(['app.py', 'test_group_resource_operation', 'get', '3'])
        self.assertEqual(dict(name="3", value="3"), json.loads(output))



from primehub.utils import resource_not_found


class GroupResourceOperation(object):

    def do_list(self, query, resource_key):
        current_group = self.group_name

        results = self.request({}, query)
        for g in results['data']['me']['effectiveGroups']:
            if current_group == g['name']:
                return g[resource_key]
        return []

    def do_get(self, query, resource_key, resource_name):
        resources = self.do_list(query, resource_key)
        data = [x for x in resources if x['name'] == resource_name]
        if data:
            return data[0]
        resource_not_found(resource_key, resource_name, 'name')



from functools import wraps

from primehub.utils import reject_action
from primehub.utils.decorators import __requires_permission__, logger, __command_groups__


def has_permission_flag(cmd_args: dict):
    k = "{}.{}".format(cmd_args['module'], cmd_args['func'])
    if k in __requires_permission__:
        return True
    return False


def disable_reject_action(action):
    logger.debug('@ask_for_permission is disable, the action will pass %s', action)


reject_action_function = disable_reject_action


def enable_ask_for_permission_feature():
    global reject_action_function
    reject_action_function = reject_action


def ask_for_permission(func):
    k = "{}.{}".format(func.__module__, func.__name__)
    __requires_permission__[k] = ""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not kwargs.get('--yes-i-really-mean-it', False):
            cmd_args = None
            try:
                actions = __command_groups__[func.__module__]
                cmd_args = [x for x in actions if x['func'] == func.__name__]
            except BaseException:
                pass
            if cmd_args:
                reject_action_function(cmd_args[0]['name'])
            else:
                reject_action_function(func.__name__)

        return func(*args, **kwargs)

    return wrapper



import importlib
import sys
from argparse import ArgumentParser, Action, RawTextHelpFormatter
from typing import Text, Iterable, Optional

from primehub.utils import create_logger

PhArgGroupClass = getattr(importlib.import_module('argparse'), '_ArgumentGroup')

logger = create_logger('primehub-parser')


class PrimeHubArgParser(ArgumentParser):

    # TODO refine the output message
    def error(self, message):
        """error(message: string)

        Prints a usage message incorporating the message to stderr and
        exits.

        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """

        logger.debug('errors: %s', message)

        # NOTE: we print usage at cli.py module
        # self.print_usage(sys.stderr)
        # args = {'prog': self.prog, 'message': message}
        # self.exit(2, _('%(prog)s: error: %(message)s xx\n') % args)
        # self.exit(1, '\n' + message + '\n\n')
        self.exit(1, '')

    def format_help(self):
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)

        # description
        if self.description:
            formatter.add_text("  " + self.description)

        if hasattr(self, 'command_description'):
            formatter.add_text(self.command_description)

        action_groups = self._action_groups
        if hasattr(self, 'available_group'):
            action_groups = [self.available_group] + self._action_groups

        # positionals, optionals and user-defined groups
        for action_group in action_groups:
            # hiding original positional arguments
            if action_group.title == 'positional arguments':
                continue
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()

    def add_command_group(self, *args, **kwargs):
        if not hasattr(self, 'available_group'):
            # add dummy command group
            available_group = PhArgGroupClass(ArgumentParser())
            available_group.title = 'Available Commands'
            self.available_group = available_group
        self.available_group.add_argument(*args, **kwargs)


class PrimeHubHelpFormatter(RawTextHelpFormatter):

    def __init__(self,
                 prog,
                 indent_increment=2,
                 max_help_position=24,
                 width=None):
        super(PrimeHubHelpFormatter, self).__init__(prog, indent_increment, max_help_position, width=120)

    def add_usage(self, usage: Text, actions: Iterable[Action], groups: Iterable[PhArgGroupClass],
                  prefix: Optional[Text] = ...) -> None:
        return super(PrimeHubHelpFormatter, self).add_usage(usage, actions, groups, "Usage: \n  ")

    def start_section(self, heading: Optional[Text]) -> None:
        if heading == "optional arguments":
            heading = "Options"
        super(PrimeHubHelpFormatter, self).start_section(heading)


def add_global_options(p):
    help_opts = p.add_argument_group('Options')
    help_opts.add_argument('-h', '--help', help='Show the help', action='store_true', default=False, dest='show_help')
    opts = p.add_argument_group('Global Options')
    opts.add_argument('--config', help='Change the path of the config file (Default: ~/.primehub/config.json)')
    opts.add_argument('--endpoint', help='Override the GraphQL API endpoint')
    opts.add_argument('--token', help='Override the API Token')
    opts.add_argument('--group', help='Override the current group')
    opts.add_argument('--json', help='Output the json format (output human-friendly format by default)',
                      dest='json_output', action='store_true')


def create_command_parser(description=None):
    p = PrimeHubArgParser(formatter_class=PrimeHubHelpFormatter, add_help=False)
    p.add_argument('command')
    p.command_description = description

    add_global_options(p)
    return p


def create_action_parser(group_command, description=None):
    p = PrimeHubArgParser(formatter_class=PrimeHubHelpFormatter, add_help=False)
    p.usage = 'primehub {} <command>'.format(group_command)
    p.add_argument('command')
    p.command_description = description

    return p


if __name__ == '__main__':
    p = create_command_parser()
    p.usage = 'primehub <command>'

    # Add fake commands to the parser's help
    p.add_command_group('config', help='xd')
    p.add_command_group('groups')
    p.print_help(file=sys.stderr)



import sys
import traceback

import primehub as ph
from primehub.config import CliConfig
from primehub.utils import create_logger, PrimeHubException, PrimeHubReturnsRequiredException, ResourceNotFoundException
from primehub.utils.argparser import create_command_parser, create_action_parser
from primehub.utils.decorators import find_actions, find_action_method, find_action_info
from primehub.utils.permission import has_permission_flag, enable_ask_for_permission_feature

logger = create_logger('primehub-cli')


def attach_dev_lab(p):
    import os
    if 'true' == os.environ.get('PRIMEHUB_SDK_DEVLAB', 'false'):
        p.register_command('extras.devlab', 'DevLab', 'devlab')
        p.register_command('extras.e2e', 'E2EForBasicFunction', 'e2e')


def create_sdk():
    cfg = ph.PrimeHubConfig()
    sdk = ph.PrimeHub(cfg)
    # Note: We replace the config to CliConfig, please see details at config module
    sdk.register_command('config', 'CliConfig')
    sdk.register_command('info', 'CliInformation')

    attach_dev_lab(sdk)
    return sdk


def reconfigure_group(sdk):
    config_command: CliConfig = sdk.config
    config_command.reconfigure_group(sdk.primehub_config.group)


def create_commands(parser, sdk):
    parsers = dict()

    for command_group in sorted(sdk.commands.keys()):
        target = sdk.commands[command_group]

        # Add the action to the main parser
        parser.add_command_group(command_group, help=target.help_description())

        # Create the group parser
        p = parsers[command_group] = create_command_parser(description=target.help_description())
        p.usage = """primehub {} <command>""".format(command_group)

        for action in find_actions(target):
            name, description = action['name'], action['description']
            p.add_command_group(name, help=description)

    return parsers


def run_action_args(sdk, selected_component, sub_parsers, target, remaining_args):
    logger.debug('invoke {} with remaining_args: {}'.format(run_action_args.__name__, remaining_args))
    helper = None

    # validate the user input command and parameters
    # primehub [command]             [unchecked params]
    #          selected_component    remaining_args
    #          sub_args.command      [params]
    #                                [action] + [params]
    sub_args, params = sub_parsers[selected_component].parse_known_args(remaining_args)
    logger.debug('sub_args: {}, params: {}'.format(sub_args, params))

    try:
        action = find_action_info(target, sub_args.command)
        if not action:
            helper = sub_parsers[selected_component]
            return helper

        # build a parser for selected action
        # primehub [command]             [action] [param1] [param2] ...
        action_parser = create_action_parser(selected_component)

        # Note: DONT add-command-group for action command
        # action_parser.add_command_group(sub_args.command, help=action['description'])
        argument_names = []
        has_kwargs = False
        logger.debug('create the action parser: {}'.format(action))
        for x in action['arguments']:
            param_name, param_type, skip_if_kwargs = x
            if skip_if_kwargs:
                # we will add optional argument in the next step
                # convert it to `--<arg-name>` format
                has_kwargs = True
                continue
            action_parser.add_argument(param_name, type=param_type)
            argument_names.append(param_name)

        # @cmd with optionals
        for x in action['optionals']:
            # There is a tuple (x, y, z, ..., arg_builder) for each optional
            # We use the arg_builder to create a new argument
            arg_builder = x[-1]
            arg_builder_args = [action_parser] + list(x[0:-1])
            arg_builder(*arg_builder_args)

        # @ask_for_permission
        if has_permission_flag(action):
            action_parser.add_argument('--yes-i-really-mean-it', dest='__primehub_permission__', action="store_true")

        action_parser.usage = 'primehub {} {} {}'.format(
            selected_component, sub_args.command, " ".join(["<%s>" % x for x in argument_names]))

        if sub_args.show_help:
            logger.debug('skip doing action, because of help calling')
            helper = action_parser
            return helper

        try:
            parsed_action_args = action_parser.parse_args([sub_args.command] + params)
            logger.debug('parsed_action_args: %s', parsed_action_args)

            # invoke <command_group>.<action>(param1, param2, ...) from the register command
            action_func = action['func']
            logger.debug('find action: {}, for target: {}'.format(action_func, target))
            func = getattr(target, action_func)

            real_parameters = []
            for x in argument_names:
                real_parameters.append(getattr(parsed_action_args, x))

            kw_parameters = {}
            if has_kwargs:
                for x in action['optionals']:
                    opt_name = x[0]
                    v = getattr(parsed_action_args, opt_name)
                    if v is not None:
                        kw_parameters[opt_name] = v

            if has_kwargs:
                if hasattr(parsed_action_args,
                           '__primehub_permission__') and parsed_action_args.__primehub_permission__:
                    kw_parameters['--yes-i-really-mean-it'] = True
                logger.debug('invoke with parameters %s {%s}', real_parameters, vars(parsed_action_args))
                return_value = func(*real_parameters, **kw_parameters)
            else:
                logger.debug('invoke with parameters %s', real_parameters)
                return_value = func(*real_parameters)
            if return_value:
                display_func = getattr(target, 'display')
                display_func(action, return_value)
            else:
                if action['return_required']:
                    raise PrimeHubReturnsRequiredException()
        except SystemExit:
            return action_parser

    except AttributeError:
        logger.debug('{}'.format(sys.exc_info()))
        traceback.print_tb(tb=sys.exc_info()[-1])
        helper = sub_parsers[selected_component]
    return helper


def run_action_noargs(sdk, selected_component, sub_parsers, target, args):
    logger.debug('invoke %s', run_action_noargs.__name__)
    helper = None
    try:
        if args.show_help:
            helper = sub_parsers[selected_component]
            return helper

        action = find_action_method(target, args.command)
        logger.debug('remaining_args is empty, find an action from parent\'s command [{}] => method: {}'
                     .format(args.command, action))
        if action is None:
            helper = sub_parsers[selected_component]
            logger.debug('cannot find a handler for %s! stop and print the help', args.command)
            return helper

        # invoke <command_group>.<action>() from the register command
        func = getattr(target, action)
        return_value = func()
        logger.debug('invoked {}'.format(func))
        if return_value:
            display_func = getattr(target, 'display')
            display_func(action, return_value)
        else:
            if action['return_required']:
                raise PrimeHubReturnsRequiredException()
    except AttributeError:
        logger.debug('{}'.format(sys.exc_info()))
        helper = sub_parsers[selected_component]
    return helper


def main(sdk=None):
    main_parser = create_command_parser()
    main_parser.usage = """primehub <command>"""

    if not sdk:
        sdk = create_sdk()

    # enable @ask_for_permission for command-line
    enable_ask_for_permission_feature()

    command_groups = create_commands(main_parser, sdk)

    hide_help = False
    helper = None
    exit_normally = False
    try:
        logger.debug('start to parse {}'.format(sys.argv))
        args, remaining_args = main_parser.parse_known_args()
        command_name = args.command
        logger.debug("args: {}".format(args))
        logger.debug("remaining_args: {}".format(remaining_args))
        reconfigure_primehub_config_if_needed(args, sdk)

        # configure output format from main-parser
        sdk.json_output = args.json_output

        if command_name not in sdk.commands:
            main_parser.epilog = 'Error: "{}" command not found'.format(command_name)
            sys.exit(1)

        # the real object to execute command
        target = sdk.commands[command_name]
        logger.debug("component: {}, target: {}".format(command_name, target))

        # got show-help, print help and exit normally
        if args.show_help and not remaining_args:
            helper = command_groups[command_name]
            exit_normally = True
            sys.exit(0)

        if args.show_help and remaining_args:
            remaining_args.append('-h')

        if remaining_args:
            helper = run_action_args(sdk, command_name, command_groups, target, remaining_args)
        else:
            helper = run_action_noargs(sdk, command_name, command_groups, target, args)

        if helper:
            sys.exit(1)

    except PrimeHubReturnsRequiredException:
        logger.debug('got PrimeHubReturnsRequiredException')
        hide_help = True
        exit_normally = False
        sys.exit(1)
    except ResourceNotFoundException as e:
        logger.debug('got PrimeHubReturnsRequiredException')
        hide_help = True
        exit_normally = False
        type_name, key, key_type = e.args
        print(f'Cannot find resource type "{type_name}" with [{key_type}: {key}]', file=sdk.stderr)
        sys.exit(1)
    except PrimeHubException as e:
        hide_help = True
        exit_normally = False
        print(e, file=sdk.stderr)
        sys.exit(1)
    except SystemExit:
        if not hide_help:
            if helper:
                helper.print_help(file=sdk.stderr)
            else:
                main_parser.print_help(file=sdk.stderr)

        if exit_normally:
            sys.exit(0)
        else:
            sys.exit(1)

    # exit application normally
    sys.exit(0)


def reconfigure_primehub_config_if_needed(args, sdk):
    if args.config is not None or args.endpoint is not None or args.group is not None or args.token is not None:
        sdk.primehub_config = ph.PrimeHubConfig(
            config=args.config, endpoint=args.endpoint,
            group=args.group, token=args.token)
        reconfigure_group(sdk)


if __name__ == '__main__':
    main()



from primehub import Helpful, cmd, Module


class Me(Helpful, Module):

    @cmd(name='me', description='Get user information', return_required=True)
    def me(self) -> dict:
        """
        Get account information

        :rtype: dict
        :returns: account information
        """
        query = """
        query {
          me {
            id
            username
            firstName
            lastName
            email
            isAdmin
          }
        }
        """
        result = self.request({}, query)
        if 'data' in result and 'me' in result['data']:
            return result['data']['me']
        return result

    def help_description(self):
        return "Show user account"



import json
import os

from primehub import Helpful, Module, cmd
from primehub.utils.optionals import toggle_flag
from primehub.utils.permission import ask_for_permission
from tests import BaseTestCase


class FakeCommand(Helpful, Module):

    @cmd(name='cmd-no-args', description='action_no_args')
    def action_no_args(self):
        return dict(name='action_no_args')

    @cmd(name='cmd-arg-str', description='action_1_args')
    def action_arg_str(self, arg1: str):
        return dict(name='action_arg_str', arg1=arg1)

    @cmd(name='cmd-arg-str-int', description='action_str_int_args')
    def action_arg_str_int(self, arg1: str, arg2: int):
        return dict(name='action_arg_str_int', arg1=arg1, arg2=arg2)

    # Note: we support optional arguments with **kwargs
    @cmd(name='cmd-args-opts', description='action_optional_args', optionals=[('arg2', str), ('page', int)])
    def action_optional_args(self, arg1: str, **kwargs):
        result = dict(name='action_arg_str_int_optional', arg1=arg1)
        if 'arg2' in kwargs:
            result['arg2'] = kwargs['arg2']
        if 'page' in kwargs:
            result['page'] = kwargs['page']
        return result

    @cmd(name='cmd-only-opts', description='action_only_optionals', optionals=[('file', str)])
    def action_only_optionals(self, **kwargs):
        return kwargs

    @ask_for_permission
    @cmd(name='cmd-remove-it', description='the action acts when user say-yes')
    def action_in_danger(self, file_to_remove, **kwargs):
        os.unlink(file_to_remove)
        return dict(result=True)

    @ask_for_permission
    @cmd(name='cmd-remove-it-no-arg-type-1', description='the action acts when user say-yes', optionals=[('file', str)])
    def action_in_danger_no_arg_type_1(self, **kwargs):
        os.unlink(kwargs.get('file'))
        return dict(result=True)

    @ask_for_permission
    @cmd(name='cmd-remove-it-no-arg-type-2', description='the action acts when user say-yes')
    def action_in_danger_no_arg_type_2(self, **kwargs):
        os.unlink("test_ask_for_permission_no_arg_type2.txt")
        return dict(result=True)

    @cmd(name='recursive-toggle-flag', description='the action has a toggle flag',
         optionals=[('recursive', toggle_flag)])
    def action_toggle_flag(self, **kwargs):
        return kwargs

    @cmd(name='return-required-no-retruns', description='return none for return required', return_required=True)
    def return_required_but_no_returns(self):
        # got exit 1
        return None

    @cmd(name='return-required-with-retruns', description='return values for return required', return_required=True)
    def return_required_with_returns(self):
        # got exit 0
        return "I don't care"

    def help_description(self):
        return "help message for fake-command"


class TestCommandGroupToCommandLine(BaseTestCase):

    def setUp(self) -> None:
        super(TestCommandGroupToCommandLine, self).setUp()

        # clean commands, add the FakeCommand
        self.sdk.register_command('test_sdk_to_cli', FakeCommand)

    def fake(self) -> FakeCommand:
        return FakeCommand(self.sdk)

    def test_command_groups(self):
        output = self.cli_stderr(['app.py', '-h'])

        # verify cli_tests command group in the output
        c = [x for x in output.split('\n') if 'test_sdk_to_cli' in x]
        self.assertTrue(c and len(c) == 1, 'We should find the command-group "test" in the output')

        description = self.fake().help_description()
        self.assertTrue(description in c[0], 'We should find the help messages in the output')

    def test_action_help(self):
        output = self.cli_stderr(['app.py', 'test_sdk_to_cli', '-h'])
        self.assertTrue('action_no_args' in output)

    def test_action_no_args(self):
        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-no-args'])
        self.assertEqual(json.dumps(self.fake().action_no_args()), output.strip())

    def test_action_1_arg(self):
        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-arg-str', 'primehub-cli-with-1-arg'])
        self.assertEqual(json.dumps(self.fake().action_arg_str('primehub-cli-with-1-arg')), output.strip())

    def test_action_with_non_str_type(self):
        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-arg-str-int', 'str', '5566'])
        self.assertEqual(json.dumps(self.fake().action_arg_str_int('str', 5566)), output.strip())

    def test_action_with_optional_args(self):
        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-args-opts', 'arg1'])
        self.assertEqual(json.dumps(self.fake().action_optional_args('arg1')), output.strip())

        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-args-opts', 'arg1', '--arg2', 'arg2'])
        self.assertEqual(json.dumps(self.fake().action_optional_args('arg1', arg2='arg2')), output.strip())

        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-args-opts', 'arg1', '--page', '9527'])
        self.assertEqual(json.dumps(self.fake().action_optional_args('arg1', page=9527)), output.strip())

    def test_action_with_only_optionals(self):
        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-only-opts', '--file', 'filename'])
        self.assertEqual(json.dumps(self.fake().action_only_optionals(file='filename')), output.strip())

    def test_ask_for_permission_1_arg(self):
        file = self.tempfile()

        # the file will not be removed, because we don't have the flag `--yes-i-really-mean-it`
        self.assertTrue(os.path.exists(file))
        self.cli_stderr(['app.py', 'test_sdk_to_cli', 'cmd-remove-it', file])
        self.assertTrue(os.path.exists(file))

        # the file will be removed by the flag `--yes-i-really-mean-it`
        self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-remove-it', file, '--yes-i-really-mean-it'])
        self.assertFalse(os.path.exists(file))

    def test_ask_for_permission_no_arg_type1(self):
        file = self.tempfile()

        # the file will not be removed, because we don't have the flag `--yes-i-really-mean-it`
        self.assertTrue(os.path.exists(file))
        self.cli_stderr(['app.py', 'test_sdk_to_cli', 'cmd-remove-it-no-arg-type-1', '--file', file])
        self.assertTrue(os.path.exists(file))

        # the file will be removed by the flag `--yes-i-really-mean-it`
        self.cli_stdout(
            ['app.py', 'test_sdk_to_cli', 'cmd-remove-it-no-arg-type-1', '--file', file, '--yes-i-really-mean-it'])
        self.assertFalse(os.path.exists(file))

    def test_ask_for_permission_no_arg_type2(self):
        file_path_to_delete = "test_ask_for_permission_no_arg_type2.txt"
        with open(file_path_to_delete, "w") as fh:
            fh.write("empty")

        # the file will not be removed, because we don't have the flag `--yes-i-really-mean-it`
        self.assertTrue(os.path.exists(file_path_to_delete))
        self.cli_stderr(['app.py', 'test_sdk_to_cli', 'cmd-remove-it-no-arg-type-2'])
        self.assertTrue(os.path.exists(file_path_to_delete))

        # the file will be removed by the flag `--yes-i-really-mean-it`
        self.cli_stdout(['app.py', 'test_sdk_to_cli', 'cmd-remove-it-no-arg-type-2', '--yes-i-really-mean-it'])
        self.assertFalse(os.path.exists(file_path_to_delete))

    def test_toggle_flag(self):
        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'recursive-toggle-flag', '--recursive'])
        self.assertEqual(dict(recursive=True), json.loads(output))

        output = self.cli_stdout(['app.py', 'test_sdk_to_cli', 'recursive-toggle-flag'])
        self.assertEqual(dict(recursive=False), json.loads(output))

    def test_return_required(self):

        def invoke_cli_and_get_exit_code(argv: list):
            try:
                import sys
                from primehub import cli
                sys.argv = argv
                cli.main(sdk=self.sdk)
            except SystemExit as e:
                if e.code == 1:
                    self.assertEqual('', self.sdk.stdout.getvalue().strip())
                    self.assertEqual('', self.sdk.stderr.getvalue().strip())
                if e.code == 0:
                    self.assertNotEqual('', self.sdk.stdout.getvalue().strip())
                return e.code

        exit_code = invoke_cli_and_get_exit_code(['app.py', 'test_sdk_to_cli', 'return-required-no-retruns'])
        self.assertEqual(1, exit_code)

        exit_code = invoke_cli_and_get_exit_code(['app.py', 'test_sdk_to_cli', 'return-required-with-retruns'])
        self.assertEqual(0, exit_code)



import json
import sys
import time
from tempfile import mkstemp

from primehub import Helpful, cmd, has_data_from_stdin, Module
from primehub.utils.permission import ask_for_permission


class DevLab(Helpful, Module):

    @cmd(name='submit-case', description='submit use case')
    def read_from_stdin_or_file(self, *args, **kwargs):
        """
        primehub devlab submit-case abc -f -xd <<EOF
        {
            "instanceType": "cpu-1",
            "image": "base-notebook",
            "displayName": "test",
            "command": "echo \"test1\"\necho \"test2\"",
        }
        EOF

        :param args:
        :param kwargs:
        :return:
        """

        if has_data_from_stdin():
            print("".join(sys.stdin.readlines()))
        print(args, kwargs)

    @cmd(name='cmd', description='show internal commands')
    def print_register_table(self):
        from primehub.utils.decorators import show_debug_info
        show_debug_info()

    @cmd(name='test-query', description='test-graphql')
    def test_query(self):
        query = """
        {
          me {
            effectiveGroups {
              id
              name
              displayName
            }
          }
        }
        """
        results = self.request({}, query)
        return results['data']['me']['effectiveGroups']

    @ask_for_permission
    @cmd(name='say-yes', description='show case for @ask_for_permission')
    def ask_for_permission(self, **kwargs):
        from datetime import datetime
        s = str(datetime.now())
        with open("ask_for_permission.txt", "w") as fh:
            fh.write(s)
            fh.write("\n")
        return dict(message='create a file [ask_for_permission.txt] having the current datetime content')

    @cmd(name='regression-job-logs', description='regression')
    def regression(self):
        instance_type = [x for x in self.primehub.instancetypes.list() if "gpu" not in x][0]['id']
        scripts = r'bash -c "for i in {1..5}; do date; sleep 1; done"'
        job_spec = dict(instanceType=instance_type, image="base-notebook", displayName="test-from-cli", command=scripts)
        fd, path = mkstemp(".json")
        with open(path, "w") as fh:
            fh.write(json.dumps(job_spec))

        my_job = self.primehub.jobs.submit(file=path)
        my_id = my_job['id']
        print("job id:", my_job['id'])

        last_state = None
        while True:
            p = self.primehub.jobs.get(my_id)['phase']
            if last_state is None:
                last_state = p

            if p != last_state:
                print("Job Phase: {} -> {}".format(last_state, p))
                last_state = p

            if last_state == 'Running':
                break

            if last_state == 'Succeeded':
                break
            time.sleep(1)

        print("Logs:")
        for g in self.primehub.jobs.logs(my_id, follow=True):
            print(g)

    def help_description(self):
        return "dev-lab is used to the primehub-python-sdk development and testing"



from types import GeneratorType

from primehub import Client
from tests import BaseTestCase


class TestHttpRequestLogs(BaseTestCase):

    def setUp(self) -> None:
        super(TestHttpRequestLogs, self).setUp()

    def test_request_logs(self):
        url = 'https://github.com/InfuseAI/primehub-python-sdk'
        c = Client(self.sdk.primehub_config)

        # with follow
        g = c.request_logs(url, True, 0)
        self.assertIsInstance(g, GeneratorType)
        count = 0
        for x in g:
            count = count + 1
        self.assertTrue(count > 1, 'Generator gives lots of lines')

        # no follow
        g = c.request_logs(url, False, 0)
        self.assertIsInstance(g, GeneratorType)
        count = 0
        for x in g:
            count = count + 1
        self.assertTrue(count > 1, 'Generator gives lots of lines')



import io
import os
import re
import sys

from jinja2 import Environment, PackageLoader, select_autoescape

from primehub import PrimeHub
from primehub.cli import create_sdk, main as cli_main
from primehub.utils.decorators import find_actions, find_action_method

env = Environment(
    loader=PackageLoader("primehub.extras"),
    autoescape=select_autoescape()
)


def get_example(command):
    try:
        return env.get_template('examples/{}.md'.format(command)).render()
    except BaseException:
        pass
    return "TBD: please write example for [{}]".format(command)


def get_doc_path():
    import primehub

    p = os.path.abspath(os.path.dirname(primehub.__file__) + "/../docs")
    return p


def create_cli_doc_path(name):
    doc_path = os.path.join(get_doc_path(), 'CLI', name + ".md")
    os.makedirs(os.path.dirname(doc_path), exist_ok=True)
    return doc_path


def generate_command_document(*args, **kwargs):
    return env.get_template('cli.tpl.md').render(*args, **kwargs)


def generate_help_for_command(sdk: PrimeHub, name):
    sdk.stderr = io.StringIO()
    sdk.stdout = io.StringIO()
    sys.argv = ['primehub', name, '-h']
    try:
        cli_main(sdk=sdk)
    except SystemExit:
        pass
    command_help = sdk.stderr.getvalue()
    actions = find_actions(sdk.commands[name])
    attach_template_information_to_action(actions, name, sdk)
    document = generate_command_document(command=name, command_help=command_help,
                                         actions=actions, examples=get_example(name))

    print("Generate doc", name)
    p = create_cli_doc_path(name)
    with open(p, "w") as fh:
        fh.write(document)


def attach_template_information_to_action(actions, name, sdk):
    for action in actions:
        explain = extract_description_from_docstring(action, name, sdk)

        def explained(x):
            if x in explain:
                return "{}: {}".format(x, explain[x])
            return x

        # arguments
        arg_list = []
        for x in action['arguments']:
            if x[2] is True:
                # skip **kwargs
                continue
            arg_list.append(x[0])
        action['required_arguments_string'] = " ".join(["<%s>" % x for x in arg_list])
        action['required_arguments'] = [explained(x) for x in arg_list]

        # optionals
        opt_list = []
        for x in action['optionals']:
            opt_list.append(x[0])
        action['optional_arguments'] = [explained(x) for x in opt_list]


def extract_description_from_docstring(action, name, sdk):
    output = dict()
    method_name = find_action_method(sdk.commands[name], action['name'])
    doc_string = getattr(sdk.commands[name], method_name).__doc__
    param_description = re.findall(r':param ([^:]+):(.+)', str(doc_string))
    for k, v in param_description:
        output[k.strip()] = v.strip()
    return output


def main():
    sdk = create_sdk()
    for k, v in sdk.commands.items():
        if k == 'devlab':
            continue
        if k == 'version':
            continue
        generate_help_for_command(sdk, k)


if __name__ == '__main__':
    main()



import inspect
import os

from primehub.utils.decorators import __requires_permission__
from tests import BaseTestCase


class TestAskForPermissionMethods(BaseTestCase):

    def test_verify_ask_for_permissions(self):
        commands = __requires_permission__.keys()
        for cmd in commands:
            if "test" in cmd:
                continue
            # cmd will look like 'primehub.schedules.delete'
            package_name, command_name, method_name = cmd.split(".")

            if package_name != 'primehub':
                # only cares about primehub package
                continue

            cmd_obj = self.sdk.commands[command_name]
            method = getattr(cmd_obj, method_name)

            # check the "**kwargs" in the last parameter
            signature = inspect.signature(method)
            signature = [str(x) for x in signature.parameters.values()]

            if not signature:
                self.report_bad_method_signature(cmd, method)

            if signature[-1] != '**kwargs':
                self.report_bad_method_signature(cmd, method)

    def report_bad_method_signature(self, cmd, method):
        src, lines = inspect.getsourcelines(method)
        delimeter = '.' * 80
        report = f'\n{delimeter} \nCommand: {cmd} \nLine: {lines}\n{"".join(src[:5])}\n{delimeter}'
        raise ValueError(
            f'@ask_for_permission must have a "**kwargs" in the last parameter, please update: {cmd}{report}')


class TestDocumentation(BaseTestCase):
    """
    There should be 3 document files for a command.

    A notebook:
    ./docs/notebook/{command}.ipynb

    A generated command line doc:
    ./docs/CLI/{command}.md

    An example will embedded into cli doc:
    ./primehub/extras/templates/examples/{command}.ipynb
    """

    def test_verify_ask_for_permissions(self):
        project_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '..'))

        def generate_doc_paths(command):
            files = [
                f'./docs/notebook/{command}.ipynb',
                f'./docs/CLI/{command}.md',
                f'./primehub/extras/templates/examples/{command}.md'
            ]
            return [os.path.normpath(os.path.join(project_dir, x)) for x in files]

        for command in self.sdk.commands:

            # skip version command
            if command == 'version':
                continue

            notebook, cli, cli_example = generate_doc_paths(command)

            if not os.path.exists(notebook):
                raise ValueError(f'There should have a {notebook} for command "{command}"')

            if not os.path.exists(cli):
                raise ValueError(f'There should have a {cli} for command "{command}"')

            if not os.path.exists(cli_example):
                raise ValueError(f'There should have a {cli_example} for command "{command}"')



#!/usr/bin/env python
import os
from distutils.core import setup

from setuptools import find_packages  # type: ignore


def _get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'primehub', 'VERSION'))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


setup(name='primehub-python-sdk',
      version=_get_version(),
      description='PrimeHub Python SDK',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      author='qrtt1',
      author_email='qrtt1@infuseai.io',
      url='https://github.com/InfuseAI/primehub-python-sdk',
      entry_points={
          'console_scripts': ['primehub = primehub.cli:main', 'doc-primehub = primehub.extras.doc_generator:main']
      },
      python_requires=">=3.6",
      packages=find_packages(),
      install_requires=['requests', 'tabulate==0.8.9', 'types-tabulate==0.8.2'],
      extras_require={
          'dev': [
              'pytest>=4.6',
              'pytest-flake8',
              'pytest-mypy',
              'pytest-cov',
              'Jinja2',
              'twine'
          ],
      },
      project_urls={
          "Bug Tracker": "https://github.com/InfuseAI/primehub/issues",
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Development Status :: 4 - Beta"
      ],
      package_data={
          'primehub': ['*.json', 'VERSION']
      })



from typing import Optional

from primehub import Helpful, cmd, Module
from primehub.resource_operations import GroupResourceOperation


class InstanceTypes(Helpful, Module, GroupResourceOperation):
    """
    List instance types or get an instance type from the list
    """
    resource_name = 'instanceTypes'
    query = """
    query {
      me {
        effectiveGroups {
          name
          instanceTypes {
            id
            name
            displayName
            description
          }
        }
      }
    }
    """

    @cmd(name='list', description='List instance types')
    def list(self) -> list:
        """
        List instance types

        :rtype: list
        :returns: all instance types in the current group
        """
        return self.do_list(InstanceTypes.query, InstanceTypes.resource_name)

    @cmd(name='get', description='Get an instance type by name', return_required=True)
    def get(self, name) -> Optional[dict]:
        """
        Get an instance type from the current group

        :type name: str
        :param name: the name of an instance type

        :rtype: Optional[dict]
        :returns: an instance type
        """
        return self.do_get(InstanceTypes.query, InstanceTypes.resource_name, name)

    def help_description(self):
        return "Get an instance types of list instance types"



from primehub import Helpful, cmd, Module


class CliInformation(Helpful, Module):

    @cmd(name='info', description='Show PrimeHub Cli information')
    def info(self):
        me = self.primehub.me.me()
        me['user_id'] = me['id']

        current_group = self.primehub.groups.get(self.group_name)
        images = [x['name'] for x in self.primehub.images.list()]
        instance_types = [x['name'] for x in self.primehub.instancetypes.list()]
        datasets = [x['name'] for x in self.primehub.datasets.list()]

        if not current_group:
            group_status = " (No matched group for name %s)" % self.group_name
        else:
            group_status = """
  Id: %(id)s
  Name: %(name)s
  Display Name: %(displayName)s
  Group Quota:
    CPU: %(projectQuotaCpu)s
    GPU: %(projectQuotaGpu)s
    Memory: %(projectQuotaMemory)s
  User Quota:
    CPU: %(quotaCpu)s
    GPU: %(quotaGpu)s
    Memory: %(quotaMemory)s
            """ % current_group

        def indent2(lines):
            return "\n  ".join(lines)

        args = dict(endpoint=self.endpoint, group_status=group_status.strip(),
                    images=indent2(images), instance_types=indent2(instance_types), datasets=indent2(datasets))
        output = """Endpoint: %(endpoint)s
User:
  Id: %(user_id)s
  Username: %(username)s
  Email: %(email)s
  First Name: %(firstName)s
  Last Name: %(lastName)s
  Is Admin: %(isAdmin)s
Current Group:
  %(group_status)s
Images:
  %(images)s
InstanceTypes:
  %(instance_types)s
Datasets:
  %(datasets)s
""" % ({**me, **args})
        return output

    def help_description(self):
        return "Display the user information and the selected group information"



import json
import os
import time
from typing import Iterator, Any

from primehub import Helpful, cmd, Module, primehub_load_config
from primehub.utils import resource_not_found, PrimeHubException
from primehub.utils.optionals import toggle_flag, file_flag


def _error_handler(response):
    import re

    if 'errors' in response:
        message = [x for x in response['errors'] if 'message' in x]
        if message:
            message = message[0]['message']
            result = re.findall(r'phjobs.primehub.io "([^"]+)" not found', message)
            if result:
                resource_not_found('job', result[0], 'id')


def invalid_config(message: str):
    example = """
    {"instanceType":"cpu-1","image":"base-notebook","displayName":"job-example","command":"echo 'good job'"}
    """.strip()
    raise PrimeHubException(message + "\n\nExample:\n" + json.dumps(json.loads(example), indent=2))


class Jobs(Helpful, Module):
    """
    The jobs module provides functions to manage PrimeHub Jobs

    Job configuration example:
    {
        "instanceType": "cpu-1",
        "image": "base-notebook",
        "displayName": "test",
        "command": "echo \"test\""
    }
    """

    def _verify_dependency(self, config):
        if 'instanceType' in config:
            self.primehub.instancetypes.get(config['instanceType'])

    @cmd(name='list', description='List jobs', optionals=[('page', int)])
    def list(self, **kwargs) -> Iterator[dict]:
        """
        List all job information in the current group

        :type page: int
        :param page: The page number as you can see in PrimeHub Jobs UI

        :rtype list
        :return The list of jobs
        """
        query = """
        query ($where: PhJobWhereInput, $page: Int, $orderBy: PhJobOrderByInput) {
          phJobsConnection(where: $where, page: $page, orderBy: $orderBy) {
            pageInfo {
              totalPage
              currentPage
            }
            edges {
              cursor
              node {
                id
                displayName
                cancel
                command
                groupId
                groupName
                schedule
                image
                instanceType {
                  id
                  name
                  displayName
                  cpuLimit
                  memoryLimit
                  gpuLimit
                }
                userId
                userName
                phase
                reason
                message
                createTime
                startTime
                finishTime
              }
            }
          }
        }
        """
        variables = {'where': {'groupId_in': [self.group_id]}, 'page': 1}
        page = kwargs.get('page', 0)
        if page:
            variables['page'] = page
            results = self.request(variables, query)
            for e in results['data']['phJobsConnection']['edges']:
                yield e['node']
            return

        page = 1
        while True:
            variables['page'] = page
            results = self.request(variables, query)
            if results['data']['phJobsConnection']['edges']:
                for e in results['data']['phJobsConnection']['edges']:
                    yield e['node']
                page = page + 1
            else:
                break

    @cmd(name='get', description='Get a job by id', return_required=True)
    def get(self, id):
        """
        Get detail information of a job by id

        :type id: str
        :param id: The job id

        :rtype dict
        :return The detail information of a job
        """
        query = """
        query ($where: PhJobWhereUniqueInput!) {
          phJob(where: $where) {
            id
            displayName
            cancel
            command
            groupId
            groupName
            schedule
            image
            instanceType {
              id
              name
              displayName
              cpuLimit
              memoryLimit
              gpuLimit
            }
            userId
            userName
            phase
            reason
            message
            createTime
            startTime
            finishTime
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['phJob']

    @cmd(name='submit', description='Submit a job', optionals=[('file', file_flag), ('from', str)])
    def _submit_cmd(self, **kwargs):
        """
        Submit a job from commands

        :type file: str
        :param file: The file path of job configurations

        :type from: str
        :param from: The schedule id to submit as a job

        :rtype dict
        :return The detail information of the submitted job
        """
        if kwargs.get('from', None):
            return self.submit_from_schedule(kwargs['from'])

        config = primehub_load_config(filename=kwargs.get('file', None))
        if not config:
            invalid_config('Job description is required.')

        config['groupId'] = self.group_id
        return self.submit(config)

    def submit(self, config):
        """
        Submit a job with config

        :type config: dict
        :param config: The job config

        :rtype dict
        :return The detail information of the submitted job
        """
        query = """
        mutation ($data: PhJobCreateInput!) {
          createPhJob(data: $data) {
            id
            displayName
            cancel
            command
            groupId
            groupName
            schedule
            image
            instanceType {
              id
              name
              displayName
              cpuLimit
              memoryLimit
              gpuLimit
            }
            userId
            userName
            phase
            reason
            message
            createTime
            startTime
            finishTime
          }
        }
        """

        if not config or (len(config) == 1):
            raise PrimeHubException('config is required')
        config['groupId'] = self.group_id

        # verify required fields in the config
        if 'instanceType' not in config:
            invalid_config('instanceType is required')
        if 'image' not in config:
            invalid_config('image is required')
        if 'displayName' not in config:
            invalid_config('displayName is required')
        if 'command' not in config:
            invalid_config('command is required')

        self._verify_dependency(config)
        results = self.request({'data': config}, query)
        return results['data']['createPhJob']

    def submit_from_schedule(self, id):
        """
        Submit a job from schedules

        :type id: str
        :param id: The schedule id

        :rtype dict
        :return The detail information of the submitted job
        """
        query = """
        mutation ($where: PhScheduleWhereUniqueInput!) {
          runPhSchedule(where: $where) {
            job {
              id
            }
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['runPhSchedule']

    @cmd(name='rerun', description='Rerun a job by id')
    def rerun(self, id):
        """
        Rerun a job by id

        :type id: str
        :param id: The job id

        :rtype dict
        :return The detail information of the ruran job
        """
        query = """
        mutation ($where: PhJobWhereUniqueInput!) {
          rerunPhJob(where: $where) {
            id
            displayName
            cancel
            command
            groupId
            groupName
            schedule
            image
            instanceType {
              id
              name
              displayName
              cpuLimit
              memoryLimit
              gpuLimit
            }
            userId
            userName
            phase
            reason
            message
            createTime
            startTime
            finishTime
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['rerunPhJob']

    @cmd(name='cancel', description='Cancel a job by id')
    def cancel(self, id):
        """
        Cancle a job by id

        :type id: str
        :param id: The job id

        :rtype dict
        :return The detail information of the canceled job
        """
        query = """
        mutation ($where: PhJobWhereUniqueInput!) {
          cancelPhJob(where: $where) {
            id
          }
        }
        """
        self.request({'where': {'id': id}}, query, _error_handler)
        return self.get(id)

    @cmd(name='wait', description='Wait a job by id', optionals=[('timeout', int)])
    def wait(self, id, **kwargs):
        """
        Wait a job in a terminated state {Succeeded, Failed, Cancelled} or until timeout

        :type id: str
        :param id: The job id

        :type timeout: int
        :param timeout: The timeout in second

        :rtype dict
        :return The detail information of the job
        """
        query = """
        query ($where: PhJobWhereUniqueInput!) {
          phJob(where: $where) {
            phase
          }
        }
        """
        timeout = kwargs.get('timeout', 0)
        start_time = time.time()
        while True:
            results = self.request({'where': {'id': id}}, query, _error_handler)
            phase = results['data']['phJob']['phase']
            if phase in ['Succeeded', 'Failed', 'Cancelled']:
                break
            time.sleep(1)
            if timeout != 0 and (time.time() - start_time >= timeout):
                break
        return self.get(id)

    @cmd(name='logs', description='Get job logs by id', optionals=[('follow', toggle_flag), ('tail', int)])
    def logs(self, id, **kwargs) -> Iterator[bytes]:
        """
        Get logs of a job

        :type id: str
        :param id: The job id

        :type follow: bool
        :param follow: Wait for additional logs to be appended

        :type tail: int
        :param tail: Show last n lines

        :rtype str
        :return The job logs
        """
        query = """
        query ($where: PhJobWhereUniqueInput!) {
          phJob(where: $where) {
            logEndpoint
          }
        }
        """

        follow = kwargs.get('follow', False)
        tail = kwargs.get('tail', 10)

        results = self.primehub.request({'where': {'id': id}}, query, _error_handler)
        endpoint = results['data']['phJob']['logEndpoint']
        return self.primehub.request_logs(endpoint, follow, tail)

    @cmd(name='list-artifacts', description='List artifacts of a job by id')
    def list_artifacts(self, id):
        """
        List all artifacts of a job

        :type id: str
        :param id: The job id

        :rtype dict
        :return The detail information of the job artifacts
        """
        query = """
        query ($where: PhJobWhereUniqueInput!) {
          phJob(where: $where) {
            artifact {
              items {
                name
                size
                lastModified
              }
            }
          }
        }
        """
        results = self.request({'where': {'id': id}}, query, _error_handler)
        return results['data']['phJob']['artifact']['items']

    @cmd(name='download-artifacts', description='Download artifacts', optionals=[('recursive', toggle_flag)])
    def download_artifacts(self, id, path, dest, **kwargs):
        """
        Download job artifacts

        :type id: str
        :param id: The job id

        :type path: str
        :param path: The path of job artifacts

        :type dest: str
        :param dest: The local path to save artifacts

        :type recusive: bool
        :param recusive: Copy recursively
        """

        if path in ['.', '', './']:
            path = '/'
        path = os.path.join('/jobArtifacts', id, path.lstrip('/'))

        # get id to verify existing
        self.get(id)
        self.primehub.files.download(path, dest, **kwargs)
        return

    def display(self, action: dict, value: Any):
        if action['func'] == 'list_artifacts' and isinstance(value, dict) and self.get_display().name != 'json':
            file_list = value.get('items', [])
            self.get_display().display(action, file_list, self.primehub.stdout)
        else:
            super(Jobs, self).display(action, value)

    def help_description(self):
        return "Get a job or list jobs"



import json
from json import JSONDecodeError
from typing import Iterator, Callable

import requests  # type: ignore

from primehub.utils import ResponseException, RequestException, GraphQLException, create_logger, \
    ResourceNotFoundException

logger = create_logger('http')


class Client(object):

    def __init__(self, primehub_config):
        self.primehub_config = primehub_config
        self.timeout = 10

    def request(self, variables: dict, query: str, error_handler: Callable = None):
        request_body = dict(variables=json.dumps(variables), query=query)
        logger.debug('request body: {}'.format(request_body))
        headers = {'authorization': 'Bearer {}'.format(self.primehub_config.api_token)}
        try:
            content = requests.post(self.primehub_config.endpoint, data=request_body, headers=headers,
                                    timeout=self.timeout).text
            logger.debug('response: {}'.format(content))
            result = json.loads(content)
            if 'errors' in result:
                if error_handler:
                    error_handler(result)
                raise GraphQLException(result)
            return result
        except JSONDecodeError:
            raise ResponseException("Response is not valid JSON:\n{}".format(content))
        except ResourceNotFoundException as e:
            raise e
        except BaseException as e:
            raise RequestException(e)

    def request_logs(self, endpoint, follow, tail) -> Iterator[bytes]:
        params = {'follow': 'false'}
        if follow:
            params['follow'] = 'true'
        if tail:
            params['tailLines'] = str(tail)
        headers = {'authorization': 'Bearer {}'.format(self.primehub_config.api_token)}

        with requests.get(endpoint, headers=headers, params=params, stream=follow) as response:
            for chunk in response.iter_content(chunk_size=8192):
                yield chunk

    def request_file(self, endpoint, dest):
        headers = {'authorization': 'Bearer {}'.format(self.primehub_config.api_token)}
        with requests.get(endpoint, headers=headers) as r:
            with open(dest, 'wb') as f:
                f.write(r.content)
        return


if __name__ == '__main__':
    print(Client.__module__)



import json
from tests import BaseTestCase


class TestCmdMe(BaseTestCase):
    """
    Usage:
      primehub me [command]

    Show user account

    Available Commands:
      me                   Get user information
    """

    def setUp(self) -> None:
        super(TestCmdMe, self).setUp()

    def test_me(self):
        # me will return anything from the query
        # query {
        #   me {
        #     id
        #     username
        #     firstName
        #     lastName
        #     email
        #     isAdmin
        #   }
        # }
        body = {'id': 'id', 'username': 'username',
                'firstName': 'firstName', 'lastName': 'lastName',
                'email': 'email', 'isAdmin': True}
        self.mock_request.return_value = {'data': {
            'me': body}}
        output = self.cli_stdout(['app.py', 'me'])

        self.assertEqual(body, json.loads(output))
